@@ -1,14 +1,16 @@
 import ast
-import textwrap
-import inspect
-import astor
-import _kratos
-from .util import print_src
-from _kratos import get_frame_local
 import copy
-import os
 import enum
+import inspect
+import os
 import sys
+import textwrap
+
+import _kratos
+import astor
+from _kratos import get_frame_local, StatementBlockType
+
+from .util import print_src
 
 
 def has_format_string():
@@ -82,7 +84,7 @@ class LogicOperatorVisitor(ast.NodeTransformer):
         return node
 
 
-class StaticElaborationNodeVisitor(ast.NodeTransformer):
+class StaticElaborationNodeForVisitor(ast.NodeTransformer):
     class NameVisitor(ast.NodeTransformer):
         def __init__(self, target, value):
             self.target = target
@@ -122,16 +124,18 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
             self.local_env[target.id] = 0
             self.global_env = global_env
 
-        def visit_Subscript(self, node: ast.Index):
+        def visit_Subscript(self, node: ast.Subscript):
             if not self.legal:
                 return
             s = node.slice
-            has_var = StaticElaborationNodeVisitor.HasVar(self.target.id)
+            has_var = StaticElaborationNodeForVisitor.HasVar(self.target.id)
             has_var.visit(s)
             if has_var.has_target:
-                if not isinstance(s, ast.Index):
-                    self.legal = False
-                    return
+                # python changed its syntax ast
+                if sys.version_info.minor < 9:
+                    if not isinstance(s, ast.Index):
+                        self.legal = False
+                        return
                 # make sure that the value is a kratos var
                 value_src = astor.to_source(node.value)
                 try:
@@ -172,7 +176,7 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
         valid = True
         for node in nodes:
             if valid:
-                visitor = StaticElaborationNodeVisitor.LoopIndexVisitor(target, self.scope, self.local, self.global_)
+                visitor = StaticElaborationNodeForVisitor.LoopIndexVisitor(target, self.scope, self.local, self.global_)
                 visitor.visit(node)
                 valid = visitor.legal
         return valid
@@ -233,7 +237,7 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
                                 args=[ast.Str(s=target.id)] + _vars,
                                 keywords=keywords,
                                 ctx=ast.Load())
-            body_visitor = StaticElaborationNodeVisitor.NameVisitor(target, index)
+            body_visitor = StaticElaborationNodeForVisitor.NameVisitor(target, index)
             body = []
             for i in range(len(node.body)):
                 n = self.visit(body_visitor.visit(node.body[i]))
@@ -253,7 +257,7 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
                 loop_body = copy.deepcopy(node.body)
                 for n in loop_body:
                     # need to replace all the reference to
-                    visitor = StaticElaborationNodeVisitor.NameVisitor(target, value)
+                    visitor = StaticElaborationNodeForVisitor.NameVisitor(target, value)
                     n = visitor.visit(n)
                     self.key_pair.append((target.id, value))
                     n = self.visit(n)
@@ -268,6 +272,26 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
                         self.target_node[n] = (target.id, value)
             return new_node
 
+
+class StaticElaborationNodeIfVisitor(ast.NodeTransformer):
+    def __init__(self, generator, fn_src, scope, local, global_, filename, func_ln):
+        super().__init__()
+        self.generator = generator
+        self.fn_src = fn_src
+        self.local = local.copy()
+        self.global_ = global_.copy()
+        self.local["self"] = self.generator
+        if "scope" in local and (not isinstance(local["scope"], Scope)):
+            raise SyntaxError("scope is a reserved keyword and shall not be used in a kratos program")
+        self.local["scope"] = scope
+        self.target_node = {}
+        self.scope = scope
+
+        self.filename = filename
+        self.scope_ln = func_ln
+
+        self.key_pair = []
+
     def __change_if_predicate(self, node):
         if isinstance(node, ast.UnaryOp):
             # notice that if the user uses `not var`, due to Python
@@ -326,14 +350,14 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
                 raise SyntaxError("Cannot statically evaluate if predicate")
             if predicate_value:
                 for i, n in enumerate(node.body):
-                    if_exp = StaticElaborationNodeVisitor(self.generator, self.fn_src, self.scope, self.unroll_for,
-                                                          self.local, self.global_, self.filename, self.scope_ln)
+                    if_exp = StaticElaborationNodeIfVisitor(self.generator, self.fn_src, self.scope,
+                                                            self.local, self.global_, self.filename, self.scope_ln)
                     node.body[i] = if_exp.visit(n)
                 return node.body
             else:
                 for i, n in enumerate(node.orelse):
-                    if_exp = StaticElaborationNodeVisitor(self.generator, self.fn_src, self.scope, self.unroll_for,
-                                                          self.local, self.global_, self.filename, self.scope_ln)
+                    if_exp = StaticElaborationNodeIfVisitor(self.generator, self.fn_src, self.scope,
+                                                            self.local, self.global_, self.filename, self.scope_ln)
                     node.orelse[i] = if_exp.visit(n)
                 return node.orelse
         else:
@@ -374,7 +398,8 @@ class StaticElaborationNodeVisitor(ast.NodeTransformer):
         else_node = ast.Call(func=ast.Attribute(attr="else_", value=if_node,
                                                 ctx=ast.Load()),
                              args=else_expression, keywords=keywords_else)
-
+        # store the mapping
+        self.target_node[node] = else_node
         return self.visit(ast.Expr(value=else_node))
 
 
@@ -392,6 +417,7 @@ class AssignNodeVisitor(ast.NodeTransformer):
         super().__init__()
         self.generator = generator
         self.debug = debug
+        self.target_node = {}
 
     def visit_Assign(self, node):
         if len(node.targets) > 1:
@@ -400,14 +426,17 @@ class AssignNodeVisitor(ast.NodeTransformer):
         args = node.targets[:] + [node.value]
         if self.debug:
             args.append(ast.Constant(value=node.lineno))
-        return ast.Expr(
+        new_node = ast.Expr(
             value=ast.Call(func=ast.Attribute(
                 value=ast.Name(id="scope",
                                ctx=ast.Load()),
                 attr="assign",
                 ctx=ast.Load()),
                 args=args,
-                keywords=[]))
+                keywords=[],
+                lineno=node.lineno))
+        self.target_node[node] = new_node
+        return new_node
 
 
 class AssertNodeVisitor(ast.NodeTransformer):
@@ -427,7 +456,8 @@ class AssertNodeVisitor(ast.NodeTransformer):
                 attr="assert_",
                 ctx=ast.Load()),
                 args=args,
-                keywords=[])
+                keywords=[],
+                lineno=node.lineno)
         return node
 
 
@@ -451,7 +481,7 @@ class ExceptionNodeVisitor(ast.NodeTransformer):
         if self.debug:
             args.append(ast.Constant(value=node.lineno))
         return ast.Call(func=ast.Attribute(value=ast.Name(id="scope", ctx=ast.Load()),
-                                           attr="assert_", ctx=ast.Load()), args=args, keywords=[])
+                                           attr="assert_", ctx=ast.Load()), args=args, keywords=[], lineno=node.lineno)
 
 
 class ReturnNodeVisitor(ast.NodeTransformer):
@@ -471,7 +501,9 @@ class ReturnNodeVisitor(ast.NodeTransformer):
             attr="return_",
             ctx=ast.Load()),
             args=args,
-            keywords=[]))
+            keywords=[],
+            lineno=node.lineno),
+            lineno=node.lineno)
 
 
 class GenVarLocalVisitor(ast.NodeTransformer):
@@ -491,7 +523,189 @@ class GenVarLocalVisitor(ast.NodeTransformer):
                                       value=ast.Str(s=self.value))
                 node.keywords.append(keyword)
 
-        return node
+        return self.generic_visit(node)
+
+
+def transform_event(ast_tree, debug, fn, ln):
+    # need to transform transaction @ event syntax into the one that supports fn ln
+    class ChangeMatMult(ast.NodeTransformer):
+        def visit_BinOp(self, node: ast.BinOp):
+            if not isinstance(node.op, ast.MatMult):
+                return node
+            # replace it with a function call that calls belongs() and set filename and line number
+            if debug:
+                args = [node.left, ast.Constant(value=fn), ast.Constant(value=ln + node.lineno - 1)]
+            else:
+                args = [node.left]
+            return ast.Expr(
+                value=ast.Call(func=ast.Attribute(attr="belongs", value=node.right, ctx=ast.Load()), args=args,
+                               ctx=ast.Load(), keywords=[]),
+                lineno=node.lineno)
+    visitor = ChangeMatMult()
+    return visitor.visit(ast_tree)
+
+
+def transform_always_comb_ssa(ast_tree, gen, _locals):
+    class SSAScope:
+        def __init__(self, var_ref):
+            self.var_ref = var_ref.copy()
+            self.created_vars = {}
+            self.enable_cond = []
+
+    class SSAVisitor(ast.NodeTransformer):
+        # https://github.com/usagitoneko97/python-static-code-analysis/tree/master/cfg_and_ssa
+        def __init__(self):
+            self.vars = set()
+            self.var_ref = {}
+
+            self.phi_scope = [SSAScope({})]
+
+        def visit_Assign(self, node):
+            node.value = self.visit(node.value)
+            # see if there is any target we need to be careful
+            assert len(node.targets) == 1, "Unpacking not supported"
+            target = node.targets[0]
+            target = self.visit(target)
+            node.targets = [target]
+            return node
+
+        def create_new_var(self, name):
+            from .passes import Attribute
+            assert name in _locals, "Only local variable scope is currently supported"
+            var = _locals[name]
+            new_name = gen.internal_generator.get_unique_variable_name("", name)
+            new_var = gen.var_from_def(var, new_name)
+            # this is used to indicate the original names
+            new_var.add_attribute(Attribute.create("ssa={0}:{1}".format(name, var.name)))
+            # scope is used to indicate the hierarchy of scopes so that the backend can reconstruct
+            # the scopes
+            new_var.add_attribute(Attribute.create("ssa-scope={0}".format(len(self.phi_scope))))
+            # set the ssa enable condition
+            if len(self.phi_scope) > 0:
+                for phi in self.phi_scope:
+                    cond = phi.enable_cond
+                    for var in cond:
+                        new_var.add_attribute(Attribute.create("ssa-en={0}".format(var)))
+            _locals[new_name] = new_var
+            self.var_ref[name] = new_name
+            self.phi_scope[-1].created_vars[name] = new_name
+            return new_name
+
+        def visit_Name(self, node: ast.Name):
+            if node.id not in self.vars:
+                # whether it's a usage or an actual assign
+                if isinstance(node.ctx, ast.Load):
+                    return node
+                else:
+                    self.vars.add(node.id)
+                    # need to create a rename
+                    name = self.create_new_var(node.id)
+                    node.id = name
+                    return node
+            else:
+                if isinstance(node.ctx, ast.Store):
+                    # need to rename again
+                    name = self.create_new_var(node.id)
+                    node.id = name
+                else:
+                    # use the existing name
+                    assert node.id in self.var_ref
+                    var_name = self.var_ref[node.id]
+                    node.id = var_name
+                return node
+
+        def __find_var_def(self, name):
+            for scope in reversed(self.phi_scope):
+                if name in scope.created_vars:
+                    return scope.created_vars[name]
+            return None
+
+        @staticmethod
+        def __add_list(target, result):
+            if isinstance(result, list):
+                for e in result:
+                    target.append(e)
+            else:
+                target.append(result)
+
+        def set_enable_cond(self, node, is_else):
+            node_str = astor.to_source(node)
+            phi = self.phi_scope[-1]
+            # add it to the enable
+            if is_else:
+                node_str = "~" + node_str
+            node_str = node_str.strip()
+            phi.enable_cond.append(node_str)
+
+        def visit_If(self, node: ast.If):
+            # push the mapping into a stack
+            # need to flatten the if statement and add phi function
+            test = self.visit(node.test)
+            body = []
+            self.phi_scope.append(SSAScope(self.var_ref))
+            # compute the enable condition
+            self.set_enable_cond(test, False)
+            for stmt in node.body:
+                stmt = self.visit(stmt)
+                self.__add_list(body, stmt)
+            body_scope: SSAScope = self.phi_scope[-1]
+            # pop the last scope
+            self.phi_scope = self.phi_scope[:-1]
+
+            else_scope: SSAScope = SSAScope({})
+            if node.orelse:
+                self.phi_scope.append(SSAScope(self.var_ref))
+                else_scope = self.phi_scope[-1]
+                # compute the enable condition
+                self.set_enable_cond(test, True)
+                for stmt in node.orelse:
+                    stmt = self.visit(stmt)
+                    self.__add_list(body, stmt)
+                # pop the last scope
+                self.phi_scope = self.phi_scope[:-1]
+
+            # if there is any variable created, need to create phi functions
+            vars_ = []
+            for var in body_scope.created_vars:
+                if var not in vars_:
+                    vars_.append(var)
+            for var in else_scope.created_vars:
+                if var not in vars_:
+                    vars_.append(var)
+            for var in vars_:
+                if var in body_scope.created_vars:
+                    true_var = body_scope.created_vars[var]
+                    # both?
+                    if var in else_scope.created_vars:
+                        # create a phi node, in this, just a ternary
+                        else_var = else_scope.created_vars[var]
+                    else:
+                        # need previous scope value
+                        else_var = self.__find_var_def(var)
+                        assert else_var is not None, "Latch {0} is created!".format(var)
+                else:
+                    else_var = else_scope.created_vars[var]
+                    # need previous scope value
+                    true_var = self.__find_var_def(var)
+                    assert true_var is not None, "Latch {0} is created!".format(var)
+                n = ast.Expr(value=ast.Call(func=ast.Name(id="ternary"), args=[test, ast.Name(id=true_var),
+                                                                               ast.Name(id=else_var)],
+                                            ctx=ast.Load(), keywords=[]),
+                             lineno=test.lineno)
+                new_name = self.create_new_var(var)
+                n = ast.Assign(targets=[ast.Name(id=new_name, ctx=ast.Store())],
+                               value=n, lineno=node.lineno)
+                body.append(n)
+
+            return body
+
+    ssa = SSAVisitor()
+    ast_tree = ssa.visit(ast_tree)
+
+    # need to add all the wires for each ref
+    for v1, v2 in ssa.var_ref.items():
+        gen.wire(gen.vars[v1], gen.vars[v2], no_fn_ln=True)
+    return ast_tree
 
 
 def add_scope_context(stmt, _locals):
@@ -672,9 +886,19 @@ def add_stmt_to_scope(fn_body):
                            args=[node], keywords=[]))
 
 
+def compute_target_node(for_nodes, new_nodes):
+    result = {}
+    for origin_node, values in for_nodes.items():
+        if origin_node in new_nodes:
+            target_node = new_nodes[origin_node]
+            result[target_node] = values
+    return result
+
+
 def __ast_transform_blocks(generator, func_tree, fn_src, fn_name, scope, insert_self,
                            filename, func_ln,
-                           transform_return=False, pre_locals=None, unroll_for=False):
+                           transform_return=False, pre_locals=None, unroll_for=False,
+                           apply_ssa=False):
     # pre-compute the frames
     # we have 3 frames back
     f = inspect.currentframe().f_back.f_back.f_back
@@ -707,15 +931,30 @@ def __ast_transform_blocks(generator, func_tree, fn_src, fn_name, scope, insert_
     aug_assign_visitor = AugAssignNodeVisitor()
     fn_body = aug_assign_visitor.visit(fn_body)
 
+    # if there is a need to apply ssa
+    if apply_ssa:
+        # only elaborate for
+        for_visitor = StaticElaborationNodeForVisitor(generator, fn_src, scope, True, _locals,
+                                                      _globals, filename, func_ln)
+        fn_body = for_visitor.visit(fn_body)
+        target_nodes = for_visitor.target_node
+        fn_body = transform_always_comb_ssa(fn_body, generator, _locals)
+    else:
+        # static eval for loop and if statement
+        for_visitor = StaticElaborationNodeForVisitor(generator, fn_src, scope, unroll_for, _locals,
+                                                      _globals, filename, func_ln)
+        fn_body = for_visitor.visit(fn_body)
+        if_visitor = StaticElaborationNodeIfVisitor(generator, fn_src, scope, _locals, _globals, filename, func_ln)
+        fn_body = if_visitor.visit(fn_body)
+        target_nodes = compute_target_node(for_visitor.target_node, if_visitor.target_node)
+
     # transform assign
     assign_visitor = AssignNodeVisitor(generator, debug)
     fn_body = assign_visitor.visit(fn_body)
     ast.fix_missing_locations(fn_body)
 
-    # static eval for loop and if statement
-    static_visitor = StaticElaborationNodeVisitor(generator, fn_src, scope, unroll_for, _locals,
-                                                  _globals, filename, func_ln)
-    fn_body = static_visitor.visit(fn_body)
+    if apply_ssa:
+        target_nodes = compute_target_node(target_nodes, assign_visitor.target_node)
 
     # transform the assert_ function to get fn_ln
     assert_visitor = AssertNodeVisitor(generator, debug)
@@ -723,8 +962,10 @@ def __ast_transform_blocks(generator, func_tree, fn_src, fn_name, scope, insert_
     exception_visitor = ExceptionNodeVisitor(generator, debug)
     fn_body = exception_visitor.visit(fn_body)
 
+    # transform event
+    transform_event(fn_body, debug, filename, func_ln)
+
     # mark the local variables
-    target_nodes = static_visitor.target_node
     for node, (key, value) in target_nodes.items():
         assign_local_visitor = GenVarLocalVisitor(key, value, "scope")
         assign_local_visitor.visit(node)
@@ -750,13 +991,6 @@ def inject_import_code(code_src):
     return "\n".join([line1, line2, code_src])
 
 
-class CodeBlockType(enum.Enum):
-    Sequential = enum.auto()
-    Combinational = enum.auto()
-    Initial = enum.auto()
-    Latch = enum.auto()
-
-
 class AlwaysWrapper:
     def __init__(self, fn):
         self.fn = fn
@@ -781,7 +1015,7 @@ def filter_fn_args(fn_args):
     return args
 
 
-def transform_stmt_block(generator, fn, unroll_for=False, fn_ln=None, kargs=None):
+def transform_stmt_block(generator, fn, unroll_for=False, apply_ssa=False, fn_ln=None, kargs=None):
     if kargs is None:
         kargs = dict()
     env_kargs = dict()
@@ -828,10 +1062,13 @@ def transform_stmt_block(generator, fn, unroll_for=False, fn_ln=None, kargs=None
     # creating the scope here
     scope = Scope(generator, filename, ln, store_local)
 
+    apply_ssa = blk_type == StatementBlockType.Combinational and apply_ssa
+
     _locals, _globals = __ast_transform_blocks(generator, func_tree, fn_src,
                                                fn_name, scope,
                                                insert_self, filename, ln,
                                                unroll_for=unroll_for,
+                                               apply_ssa=apply_ssa,
                                                pre_locals=env_kargs)
 
     src = astor.to_source(func_tree, pretty_source=__pretty_source)
@@ -925,9 +1162,9 @@ def extract_arg_name_order_from_fn(fn):
 
 def extract_sensitivity_from_dec(deco_list, fn_name):
     if len(deco_list) == 0:
-        return CodeBlockType.Combinational, []
+        return StatementBlockType.Combinational, []
     else:
-        assert len(deco_list) == 1,  "{0} is not called with multiple decorators blocks".format(fn_name)
+        assert len(deco_list) == 1, "{0} is not called with multiple decorators blocks".format(fn_name)
         call_obj = deco_list[0]
         if isinstance(call_obj, ast.Call):
             call_name = call_obj.func.id
@@ -935,14 +1172,16 @@ def extract_sensitivity_from_dec(deco_list, fn_name):
             assert isinstance(call_obj, ast.Name), "Unrecognized  function decorator {0}".format(call_obj)
             call_name = call_obj.id
         if call_name == "always_comb":
-            return CodeBlockType.Combinational, []
+            return StatementBlockType.Combinational, []
         elif call_name == "initial":
-            return CodeBlockType.Initial, []
+            return StatementBlockType.Initial, []
         elif call_name == "always_latch":
-            return CodeBlockType.Latch, []
+            return StatementBlockType.Latch, []
+        elif call_name == "final":
+            return StatementBlockType.Final, []
         else:
             assert call_name == "always_ff", "Unrecognized function decorator {0}".format(call_name)
-        blk_type = CodeBlockType.Sequential
+        blk_type = StatementBlockType.Sequential
         raw_sensitivity = call_obj.args
         result = []
         # TODO: fix me. the frame num calculation is a hack
@@ -960,12 +1199,12 @@ def extract_sensitivity_from_dec(deco_list, fn_name):
                 assert name in local, "{0} not found".format(name)
                 n = eval(name, local)
                 assert isinstance(n, _kratos.Var), "{0} is not a variable".format(name)
-                signal_name = str(n)
+                signal_name = n
             elif isinstance(signal_name_node, ast.Attribute):
                 # need to eval the actual name
                 n = eval(astor.to_source(signal_name_node), local)
                 assert isinstance(n, _kratos.Var), "{0} is not a variable".format(signal_name_node)
-                signal_name = str(n)
+                signal_name = n
             else:
                 signal_name = signal_name_node.s
             result.append((edge_type, signal_name))
