@@ -9,16 +9,18 @@
 namespace py = pybind11;
 using std::shared_ptr;
 
+// some IDE uses system header, which might be python2
+#if PY_MAJOR_VERSION == 2
+extern PyObject *_PyLong_Format(PyObject *obj, int base);
+#endif
+
 std::optional<std::pair<std::string, uint32_t>> get_fn_ln(uint32_t num_frame_back);
 
 // templated function to set up packed struct for both port and var
 template <class T, class K>
-void setup_getitem(py::class_<T, ::shared_ptr<T>, K> &class_) {
+void setup_getitem(py::class_<T, ::shared_ptr<T>, K> &class_, bool add_str = true) {
     using namespace kratos;
     class_
-        .def(
-            "__getitem__", [](T & var, const std::string &name) -> auto & { return var[name]; },
-            py::return_value_policy::reference)
         .def(
             "__getitem__",
             [](T & var, uint32_t index) -> auto & { return var.Var::operator[](index); },
@@ -33,6 +35,12 @@ void setup_getitem(py::class_<T, ::shared_ptr<T>, K> &class_) {
                 return var.Var::operator[](slice);
             },
             py::return_value_policy::reference);
+
+    if (add_str) {
+        class_.def(
+            "__getitem__", [](T & var, const std::string &name) -> auto & { return var[name]; },
+            py::return_value_policy::reference);
+    }
 }
 
 void init_common_expr(py::class_<kratos::Var, ::shared_ptr<kratos::Var>> &class_) {
@@ -153,7 +161,22 @@ void init_common_expr(py::class_<kratos::Var, ::shared_ptr<kratos::Var>> &class_
         .def(
             "__rshift__",
             [](const int64_t &left, Var &right) -> Expr & {
-                return convert_int_to_const(left, right) >> right;
+                return convert_int_to_const(left, right) << right;
+            },
+            py::return_value_policy::reference)  // NOLINT
+        .def(
+            "__lshift__", [](const Var &left, const Var &right) -> Expr & { return left << right; },
+            py::return_value_policy::reference)  // NOLINT
+        .def(
+            "__lshift__",
+            [](const Var &left, const int64_t &right) -> Expr & {
+                return left << convert_int_to_const(left, right);
+            },
+            py::return_value_policy::reference)  // NOLINT
+        .def(
+            "__lshift__",
+            [](const int64_t &left, Var &right) -> Expr & {
+                return convert_int_to_const(left, right) << right;
             },
             py::return_value_policy::reference)  // NOLINT
         .def(
@@ -363,7 +386,44 @@ void init_common_expr(py::class_<kratos::Var, ::shared_ptr<kratos::Var>> &class_
         .def("type", &Var::type)
         .def("concat", &Var::concat, py::return_value_policy::reference, py::arg("var"))
         .def("extend", &Var::extend, py::return_value_policy::reference, py::arg("width"))
-        .def_readwrite("name", &Var::name)
+        .def_property(
+            "name", [](const Var &var) { return var.name; },
+            [](Var &var, const std::string &value) {
+                if (var.name == value) return;
+                auto *gen = var.generator();
+                if (gen->has_var(value)) {
+                    throw UserException(
+                        fmt::format("{0} already exists in {1}", value, gen->instance_name));
+                }
+                var.name = value;
+                if (gen->debug) {
+                    auto fn_ln = get_fn_ln(1);
+                    if (fn_ln) {
+                        var.fn_name_ln.emplace_back(*fn_ln);
+                    }
+                }
+            })
+        .def("rename",  // rename will reindex the ports. cannot be used in a pass
+             [](Var &var, const std::string &value) {
+                 // this is for pass usage only. need to reindex the variables
+                 // once you're done
+                 // maybe refactor how the IR visits vars?
+                 if (var.name == value) return;
+                 auto *gen = var.generator();
+                 if (gen->has_var(value)) {
+                     throw UserException(
+                         fmt::format("{0} already exists in {1}", value, gen->instance_name));
+                 }
+                 var.name = value;
+
+                 gen->reindex_vars();
+                 if (gen->debug) {
+                     auto fn_ln = get_fn_ln(1);
+                     if (fn_ln) {
+                         var.fn_name_ln.emplace_back(*fn_ln);
+                     }
+                 }
+             })
         .def_property(
             "width", [](Var &var) { return var.var_width(); },
             [](Var &var, uint32_t width) {
@@ -417,6 +477,38 @@ void init_getitem(py::class_<kratos::Var, ::shared_ptr<kratos::Var>> &class_) {
             py::return_value_policy::reference);
 }
 
+void init_big_int(py::module &m) {
+    m.def(
+        "constant",
+        [](const py::object &obj, uint64_t width, bool is_signed) {
+            auto num_bits = _PyLong_NumBits(obj.ptr());
+
+            if (num_bits > width)
+                throw std::invalid_argument(
+                    fmt::format("Width ({0}) too small for constant value", width));
+            if (num_bits <= 64) {
+                auto v = PyLong_AsLongLong(obj.ptr());
+                auto &c = kratos::Const::constant(v, width, is_signed);
+                return c.shared_from_this();
+            }
+            auto *format = _PyLong_Format(obj.ptr(), 16);
+            auto str_value = py::cast<std::string>(format);
+            bool negative = false;
+            if (str_value[0] == '-') {
+                if (!is_signed)
+                    throw std::invalid_argument(
+                        "Constant value is signed but argument says otherwise");
+                negative = true;
+                str_value = str_value.substr(1);
+            }
+            // remove 0x
+            str_value = str_value.substr(2);
+            auto &c = kratos::Const::constant(str_value, num_bits, negative, width, is_signed);
+            return c.shared_from_this();
+        },
+        py::return_value_policy::reference);
+}
+
 // deal with all the expr/var stuff
 void init_expr(py::module &m) {
     using namespace kratos;
@@ -428,7 +520,7 @@ void init_expr(py::module &m) {
     auto expr = py::class_<Expr, ::shared_ptr<Expr>, Var>(m, "Expr");
 
     auto port = py::class_<Port, ::shared_ptr<Port>, Var>(m, "Port");
-    port.def_property_readonly("port_direction", &Port::port_direction)
+    port.def_property("port_direction", &Port::port_direction, &Port::set_port_direction)
         .def_property("port_type", &Port::port_type, &Port::set_port_type)
         .def("connected_to", &Port::connected_to)
         .def("connected_from", &Port::connected_from)
@@ -437,6 +529,7 @@ void init_expr(py::module &m) {
 
     auto const_ = py::class_<Const, ::shared_ptr<Const>, Var>(m, "Const");
     const_.def("value", &Const::value).def("set_value", &Const::set_value);
+    const_.def_property_readonly("is_bignum", &Const::is_bignum);
 
     auto slice = py::class_<VarSlice, ::shared_ptr<VarSlice>, Var>(m, "VarSlice");
     slice.def_property_readonly("sliced_by_var", &VarSlice::sliced_by_var)
@@ -462,7 +555,8 @@ void init_expr(py::module &m) {
             [](VarSlice &var, const std::shared_ptr<Var> &slice) -> VarSlice & {
                 return var[slice];
             },
-            py::return_value_policy::reference);
+            py::return_value_policy::reference)
+        .def_readonly("parent_var", &VarSlice::parent_var, py::return_value_policy::reference);
 
     auto var_slice = py::class_<VarVarSlice, ::shared_ptr<VarVarSlice>, VarSlice>(m, "VarVarSlice");
     var_slice.def_property_readonly(
@@ -530,15 +624,43 @@ void init_expr(py::module &m) {
     setup_getitem(var_packed);
 
     // struct info for packed
-    auto struct_ = py::class_<PackedStruct>(m, "PackedStruct");
+    auto struct_ = py::class_<PackedStruct, std::shared_ptr<PackedStruct>>(m, "PackedStruct");
     struct_.def(py::init<std::string, std::vector<std::tuple<std::string, uint32_t, bool>>>())
         .def(py::init<std::string, std::vector<std::tuple<std::string, uint32_t>>>())
+        .def(py::init<std::string>())
         .def_readwrite("struct_name", &PackedStruct::struct_name)
         .def_readonly("attributes", &PackedStruct::attributes)
-        .def_readwrite("external", &PackedStruct::external);
+        .def_readwrite("external", &PackedStruct::external)
+        .def("add_attribute",
+             [](PackedStruct &struct_, const std::string &name, uint32_t width) {
+                 auto attr = PackedStructFieldDef();
+                 attr.name = name;
+                 attr.width = width;
+                 attr.signed_ = false;
+                 struct_.attributes.emplace_back(attr);
+             })
+        .def("add_attribute",
+             [](PackedStruct &struct_, const std::string &name, uint32_t width, bool signed_) {
+                 auto attr = PackedStructFieldDef();
+                 attr.name = name;
+                 attr.width = width;
+                 attr.signed_ = signed_;
+                 struct_.attributes.emplace_back(attr);
+             })
+        .def("add_attribute", [](PackedStruct &struct_, const std::string &name,
+                                 const std::shared_ptr<PackedStruct> &s) {
+            auto attr = PackedStructFieldDef();
+            attr.name = name;
+            attr.struct_ = s.get();
+            struct_.attributes.emplace_back(attr);
+        });
 
-    auto port_packed_slice =
-        py::class_<PackedSlice, ::shared_ptr<PackedSlice>, Var>(m, "PackedSlice");
+    auto packed_slice = py::class_<PackedSlice, ::shared_ptr<PackedSlice>, Var>(m, "PackedSlice");
+    packed_slice.def(
+        "__getitem__",
+        [](PackedSlice &p, const std::string &mem) -> PackedSlice & { return p.slice_member(mem); },
+        py::return_value_policy::reference);
+    setup_getitem(packed_slice, false);
 
     // ternary op
     auto ternary_exp =
@@ -551,6 +673,7 @@ void init_expr(py::module &m) {
 
     // constant
     m.def("constant", &constant, py::return_value_policy::reference);
+    init_big_int(m);
 
     m.def("mux", util::mux);
 
