@@ -17,7 +17,7 @@ using std::shared_ptr;
 void init_generator(py::module &m) {
     using namespace kratos;
     auto generator = py::class_<Generator, ::shared_ptr<Generator>, IRNode>(m, "Generator");
-    generator.def("from_verilog", &Generator::from_verilog)
+    generator.def("from_verilog", &Generator::from_verilog, py::return_value_policy::reference)
         .def("var", py::overload_cast<const std::string &, uint32_t>(&Generator::var),
              py::return_value_policy::reference)
         .def("var",
@@ -49,6 +49,8 @@ void init_generator(py::module &m) {
                 return v;
             },
             py::return_value_policy::reference)
+        .def("var", py::overload_cast<const Var &, const std::string &>(&Generator::var),
+             py::return_value_policy::reference)
         .def("port",
              py::overload_cast<PortDirection, const std::string &, uint32_t>(&Generator::port),
              py::return_value_policy::reference)
@@ -151,26 +153,31 @@ void init_generator(py::module &m) {
                  return interfaces.at(name);
              })
         .def("port_packed",
-             py::overload_cast<PortDirection, const std::string &, const PackedStruct &>(
-                 &Generator::port_packed),
+             py::overload_cast<PortDirection, const std::string &,
+                               const std::shared_ptr<PackedStruct> &>(&Generator::port_packed),
              py::return_value_policy::reference)
         .def("port_packed",
-             py::overload_cast<PortDirection, const std::string &, const PackedStruct &, uint32_t>(
+             py::overload_cast<PortDirection, const std::string &,
+                               const std::shared_ptr<PackedStruct> &, uint32_t>(
                  &Generator::port_packed),
              py::return_value_policy::reference)
-        .def("port_packed",
-             py::overload_cast<PortDirection, const std::string &, const PackedStruct &,
-                               const std::vector<uint32_t> &>(&Generator::port_packed),
-             py::return_value_policy::reference)
+        .def(
+            "port_packed",
+            py::overload_cast<PortDirection, const std::string &,
+                              const std::shared_ptr<PackedStruct> &, const std::vector<uint32_t> &>(
+                &Generator::port_packed),
+            py::return_value_policy::reference)
         .def("var_packed",
-             py::overload_cast<const std::string &, const PackedStruct &>(&Generator::var_packed),
-             py::return_value_policy::reference)
-        .def("var_packed",
-             py::overload_cast<const std::string &, const PackedStruct &, uint32_t>(
+             py::overload_cast<const std::string &, const std::shared_ptr<PackedStruct> &>(
                  &Generator::var_packed),
              py::return_value_policy::reference)
+        .def(
+            "var_packed",
+            py::overload_cast<const std::string &, const std::shared_ptr<PackedStruct> &, uint32_t>(
+                &Generator::var_packed),
+            py::return_value_policy::reference)
         .def("var_packed",
-             py::overload_cast<const std::string &, const PackedStruct &,
+             py::overload_cast<const std::string &, const std::shared_ptr<PackedStruct> &,
                                const std::vector<uint32_t> &>(&Generator::var_packed),
              py::return_value_policy::reference)
         .def("enum", &Generator::enum_, py::return_value_policy::reference)
@@ -193,6 +200,7 @@ void init_generator(py::module &m) {
         .def("sequential", &Generator::sequential, py::return_value_policy::reference)
         .def("combinational", &Generator::combinational, py::return_value_policy::reference)
         .def("initial", &Generator::initial, py::return_value_policy::reference)
+        .def("final", &Generator::final, py::return_value_policy::reference)
         .def("latch", &Generator::latch, py::return_value_policy::reference)
         .def("add_child_generator",
              py::overload_cast<const std::string &, const std::shared_ptr<Generator> &>(
@@ -215,6 +223,7 @@ void init_generator(py::module &m) {
         .def("set_is_stub", &Generator::set_is_stub)
         .def("wire_ports", &Generator::wire_ports)
         .def("wire", &Generator::wire)
+        .def("unwire", &Generator::unwire)
         .def("wire_interface", &Generator::wire_interface)
         .def("correct_wire_direction", &Generator::correct_wire_direction)
         .def("correct_wire_direction",
