@@ -73,8 +73,7 @@ Stream& Stream::operator<<(AssignStmt* stmt) {
     for (uint64_t i = 1; i < right_wrapped.size(); i++) {
         // compute new indent
         (*this) << endl();
-        auto new_indent = codegen_->indent() + "    ";
-        (*this) << new_indent << right_wrapped[i];
+        (*this) << codegen_->indent() << "    " << right_wrapped[i];
     }
     (*this) << ";" << endl();
     return *this;
@@ -99,7 +98,7 @@ std::string Stream::get_var_decl(kratos::Var* var) {
     std::string type;
     if (var->is_struct()) {
         auto v = var->as<VarPackedStruct>();
-        type = v->packed_struct().struct_name;
+        type = v->packed_struct()->struct_name;
     } else if (var->is_enum()) {
         auto* v = dynamic_cast<EnumType*>(var);
         if (!v) throw InternalException("Unable to resolve var to enum type");
@@ -366,16 +365,29 @@ void SystemVerilogCodeGen::generate_functions(kratos::Generator* generator) {
     for (auto const& iter : funcs) stmt_code(iter.second.get());
 }
 
-std::string SystemVerilogCodeGen::indent() {
+std::string_view SystemVerilogCodeGen::indent() {
     if (skip_indent_) {
         skip_indent_ = false;
         return "";
     }
-    std::string result;
-    uint32_t num_indent = indent_ * indent_size;
-    result.resize(num_indent);
-    for (uint32_t i = 0; i < num_indent; i++) result[i] = ' ';
-    return result;
+    if (empty_indent_str_.size() < (indent_ * indent_size)) {
+        // need to expand the string
+        std::string result;
+        uint32_t num_indent = indent_ * indent_size;
+        result.resize(num_indent);
+        for (uint32_t i = 0; i < num_indent; i++) result[i] = ' ';
+        empty_indent_str_ = result;
+        empty_indent_string_view_ = empty_indent_str_;
+    }
+    return empty_indent_string_view_.substr(0, indent_ * indent_size);
+}
+
+std::string_view SystemVerilogCodeGen::pre_indent() {
+    if (indent_ == 0) {
+        return "";
+    } else {
+        return empty_indent_string_view_.substr(0, (indent_ - 1) * indent_size);
+    }
 }
 
 void SystemVerilogCodeGen::dispatch_node(IRNode* node) {
@@ -388,32 +400,47 @@ void SystemVerilogCodeGen::dispatch_node(IRNode* node) {
     // yosys src
     if (yosys_src_) output_yosys_src(node);
 
-    if (stmt_ptr->type() == StatementType::Assign) {
-        stmt_code(reinterpret_cast<AssignStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::Block) {
-        stmt_code(reinterpret_cast<StmtBlock*>(node));
-    } else if (stmt_ptr->type() == StatementType::If) {
-        stmt_code(reinterpret_cast<IfStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::ModuleInstantiation) {
-        stmt_code(reinterpret_cast<ModuleInstantiationStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType ::Switch) {
-        stmt_code(reinterpret_cast<SwitchStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType ::FunctionalCall) {
-        stmt_code(reinterpret_cast<FunctionCallStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::Return) {
-        stmt_code(reinterpret_cast<ReturnStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::Assert) {
-        stmt_code(reinterpret_cast<AssertBase*>(node));
-    } else if (stmt_ptr->type() == StatementType::Comment) {
-        stmt_code(reinterpret_cast<CommentStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::InterfaceInstantiation) {
-        // do nothing
-    } else if (stmt_ptr->type() == StatementType::RawString) {
-        stmt_code(reinterpret_cast<RawStringStmt*>(node));
-    } else if (stmt_ptr->type() == StatementType::For) {
-        stmt_code(reinterpret_cast<ForStmt*>(node));
-    } else {
-        throw StmtException("Not implemented", {node});
+    // use switch for branch tables
+    // also let compiler check if we have all the types covered
+    switch (stmt_ptr->type()) {
+        case StatementType::Assign:
+            stmt_code(reinterpret_cast<AssignStmt*>(node));
+            break;
+        case StatementType::Block:
+            stmt_code(reinterpret_cast<StmtBlock*>(node));
+            break;
+        case StatementType::If:
+            stmt_code(reinterpret_cast<IfStmt*>(node));
+            break;
+        case StatementType::ModuleInstantiation:
+            stmt_code(reinterpret_cast<ModuleInstantiationStmt*>(node));
+            break;
+        case StatementType::Switch:
+            stmt_code(reinterpret_cast<SwitchStmt*>(node));
+            break;
+        case StatementType::FunctionalCall:
+            stmt_code(reinterpret_cast<FunctionCallStmt*>(node));
+            break;
+        case StatementType::Return:
+            stmt_code(reinterpret_cast<ReturnStmt*>(node));
+            break;
+        case StatementType::Assert:
+            stmt_code(reinterpret_cast<AssertBase*>(node));
+            break;
+        case StatementType::Comment:
+            stmt_code(reinterpret_cast<CommentStmt*>(node));
+            break;
+        case StatementType::InterfaceInstantiation:
+            // do nothing
+            break;
+        case StatementType::RawString:
+            stmt_code(reinterpret_cast<RawStringStmt*>(node));
+            break;
+        case StatementType::For:
+            stmt_code(reinterpret_cast<ForStmt*>(node));
+            break;
+        case StatementType::Auxiliary:
+            throw UserException("Auxiliary statement should not be in the codegen!");
     }
 }
 
@@ -457,6 +484,10 @@ void SystemVerilogCodeGen::stmt_code(StmtBlock* stmt) {
             stmt_code(reinterpret_cast<InitialStmtBlock*>(stmt));
             break;
         }
+        case StatementBlockType::Final: {
+            stmt_code(reinterpret_cast<FinalStmtBlock*>(stmt));
+            break;
+        }
         case StatementBlockType::Latch: {
             stmt_code(reinterpret_cast<LatchStmtBlock*>(stmt));
             break;
@@ -535,6 +566,26 @@ void SystemVerilogCodeGen::stmt_code(kratos::InitialStmtBlock* stmt) {
     stream_ << indent() << "end" << block_label(stmt) << stream_.endl();
 }
 
+void SystemVerilogCodeGen::stmt_code(kratos::FinalStmtBlock* stmt) {
+    // comment
+    if (!stmt->comment.empty()) {
+        stream_ << indent() << "// " << strip_newline(stmt->comment) << stream_.endl();
+    }
+    if (generator_->debug) {
+        stmt->verilog_ln = stream_.line_no();
+    }
+
+    stream_ << "final begin" << block_label(stmt) << stream_.endl();
+    indent_++;
+
+    for (uint64_t i = 0; i < stmt->size(); i++) {
+        dispatch_node(stmt->get_child(i));
+    }
+
+    indent_--;
+    stream_ << indent() << "end" << block_label(stmt) << stream_.endl();
+}
+
 void SystemVerilogCodeGen::stmt_code(kratos::ScopedStmtBlock* stmt) {
     if (generator_->debug) {
         stmt->verilog_ln = stream_.line_no();
@@ -881,16 +932,30 @@ void SystemVerilogCodeGen::stmt_code(kratos::ForStmt* stmt) {
     // for loop
     if (generator_->debug) stmt->verilog_ln = stream_.line_no();
     auto iter = stmt->get_iter_var();
-    stream_ << indent() << "for (int " << (iter->is_signed() ? " " : "unsigned ")
-            << iter->to_string() << " = ";
+
+    // get var declaration
+    std::vector<std::string> var_decl;
+    if (iter->is_gen_var()) {
+        // genvar don't have type declaration
+        var_decl.emplace_back("genvar");
+    } else {
+        // maybe add support for much bigger number?
+        var_decl.emplace_back("int");
+        if (!iter->is_signed()) var_decl.emplace_back("unsigned");
+    }
+
+    var_decl.emplace_back(iter->to_string());
+    auto var_decl_str = string::join(var_decl.begin(), var_decl.end(), " ");
+
+    stream_ << indent() << "for (" << var_decl_str << " = ";
     stream_ << ::format("{0}", stmt->start()) << "; " << iter->to_string()
             << (stmt->end() > stmt->start() ? " < " : " > ");
     stream_ << ::format("{0}", stmt->end()) << "; " << iter->to_string()
             << (stmt->step() > 0 ? " += " : " -= ");
     stream_ << ::format("{0}", std::abs(stmt->step())) << ") ";
-    indent_++;
+    if (!iter->is_gen_var()) indent_++;
     dispatch_node(stmt->get_loop_body().get());
-    indent_--;
+    if (!iter->is_gen_var()) indent_--;
 }
 
 void SystemVerilogCodeGen::stmt_code(LatchStmtBlock* stmt) { block_code("always_latch", stmt); }
@@ -923,7 +988,7 @@ std::string SystemVerilogCodeGen::get_port_str(Port* port) {
         strs.emplace_back(p->to_string());
     } else {
         auto* ptr = reinterpret_cast<PortPackedStruct*>(port);
-        strs.emplace_back(ptr->packed_struct().struct_name);
+        strs.emplace_back(ptr->packed_struct()->struct_name);
     }
     if (port->is_signed()) strs.emplace_back("signed");
     if ((port->size().front() > 1 || port->size().size() > 1 || port->explicit_array() ||
@@ -1030,8 +1095,11 @@ void SystemVerilogCodeGen::generate_port_interface(kratos::InstantiationStmt* st
                     auto mod_port_name = internal_def->definition()->name();
                     external_name = external->base_name();
                     // FIXME: this is a little bit hacky here
-                    if (external_name.find('.') == std::string::npos)
-                        external_name = ::format("{0}.{1}", external_name, mod_port_name);
+                    auto potential_name = ::format("{0}.{1}", external_name, mod_port_name);
+                    auto i = external->generator()->get_interface(external_name);
+                    if (i && !i->definition()->is_modport()) {
+                        external_name = potential_name;
+                    }
                 } else {
                     external_name = external->base_name();
                 }
@@ -1058,7 +1126,7 @@ void SystemVerilogCodeGen::generate_port_interface(kratos::InstantiationStmt* st
         else
             stream_ << stream_.endl();
     }
-    stream_ << ");" << stream_.endl() << stream_.endl();
+    stream_ << pre_indent() << ");" << stream_.endl() << stream_.endl();
     indent_--;
 }
 
@@ -1137,7 +1205,7 @@ public:
             auto p = generator->get_port(port_name);
             if (p->is_interface()) {
                 auto interface_p = p->as<InterfacePort>();
-                const auto *ref = interface_p->interface();
+                const auto* ref = interface_p->interface();
                 auto def = ref->definition();
                 update_interface_definition(def, ref, generator);
             }
@@ -1199,7 +1267,7 @@ private:
 std::map<std::string, std::string> extract_interface_info(Generator* top) {
     InterfaceVisitor visitor;
     visitor.visit_generator_root_p(top);
-    auto defs = visitor.interfaces();
+    auto const& defs = visitor.interfaces();
     std::map<std::string, std::string> result;
     const std::string indent = "  ";
     for (auto const& [interface_name, def] : defs) {
@@ -1259,6 +1327,13 @@ std::map<std::string, std::string> extract_interface_info(Generator* top) {
     return result;
 }
 
+void copy_attrs(const Var& src, Var& target) {
+    auto const& attributes = src.get_attributes();
+    for (auto const& attr : attributes) {
+        target.add_attribute(attr);
+    }
+}
+
 Generator& create_wrapper_flatten(Generator* top, const std::string& wrapper_name) {
     auto& gen = top->context()->generator(wrapper_name);
     gen.add_child_generator(top->instance_name, top->shared_from_this());
@@ -1281,6 +1356,7 @@ Generator& create_wrapper_flatten(Generator* top, const std::string& wrapper_nam
             } else {
                 gen.add_stmt(new_port.assign(p, AssignmentType::Blocking));
             }
+            copy_attrs(*p, new_port);
         } else {
             // need to flatten the array
             auto slices = get_flatten_slices(p.get());
@@ -1299,6 +1375,7 @@ Generator& create_wrapper_flatten(Generator* top, const std::string& wrapper_nam
                     gen.add_stmt(
                         new_port.assign(slice_port->shared_from_this(), AssignmentType::Blocking));
                 }
+                copy_attrs(*p, new_port);
             }
         }
     }
