@@ -2,11 +2,12 @@
 
 #include <mutex>
 
-#include "db.hh"
 #include "except.hh"
 #include "fmt/format.h"
 #include "generator.hh"
+#include "graph.hh"
 #include "pass.hh"
+#include "schema.hh"
 #include "tb.hh"
 #include "util.hh"
 
@@ -26,72 +27,34 @@ public:
 
     void visit(ScopedStmtBlock *stmt) override { insert_statements(stmt); }
 
+    void visit(AssignStmt *stmt) override {
+        auto const *parent = stmt->parent();
+        if (parent->ir_node_kind() == IRNodeKind::GeneratorKind) {
+            auto const *p = reinterpret_cast<const Generator *>(parent);
+            if (p->debug) {
+                process_stmt(stmt);
+            }
+        }
+    }
+
 private:
     void insert_statements(StmtBlock *block) {
         auto *parent = block->generator_parent();
         if (!parent->debug)
             // we only insert statements to the ones that has debug turned on
             return;
-        std::vector<std::shared_ptr<Stmt>> new_blocks;
-        new_blocks.reserve(block->size() * 2);
 
         for (auto const &stmt : *block) {
-            uint32_t stmt_id = context_->max_stmt_id()++;
-            // make a function call
-            auto bp_stmt = get_function_call_stmt(parent, stmt_id);
-            // FIXME: See #89
-            bp_stmt->set_parent(block);
-            new_blocks.emplace_back(bp_stmt);
-            // insert the normal one
-            new_blocks.emplace_back(stmt);
-            // set stmt id
-            stmt->set_stmt_id(stmt_id);
+            process_stmt(stmt.get());
         }
-
-        // replace the content
-        block->set_stmts(new_blocks);
     }
 
-    std::shared_ptr<FunctionCallStmt> get_function_call_stmt(Generator *generator,
-                                                             uint32_t stmt_id) {
-        if (!generator->has_function(break_point_func_name)) {
-            // create the function in the generator
-            auto func = generator->dpi_function(break_point_func_name);
-            func->input(break_point_instance_id_arg, var_size_, false);
-            func->input(var_name_, var_size_, false);
-            func->set_port_ordering({{break_point_instance_id_arg, 0}, {break_point_func_arg, 1}});
-        }
-        auto &id_const = constant(stmt_id, var_size_);
-        auto i_id_const = generator->get_param(break_point_param_name);
-        if (!i_id_const) {
-            throw UserException(::format("{0} not found for {1}", break_point_param_name,
-                                         generator->handle_name()));
-        }
-        auto &var = generator->call(
-            break_point_func_name,
-            {{var_name_, id_const.shared_from_this()}, {break_point_instance_id_arg, i_id_const}},
-            false);
-        return std::make_shared<FunctionCallStmt>(var.as<FunctionCallVar>());
+    void process_stmt(Stmt *stmt) {
+        uint32_t stmt_id = context_->max_stmt_id()++;
+        // set stmt id
+        stmt->set_stmt_id(stmt_id);
     }
 
-    static StmtBlock *get_root_block(Stmt *stmt) {
-        if (stmt->type() == StatementType::Block) {
-            auto *block = reinterpret_cast<StmtBlock *>(stmt);
-            if (block->block_type() != StatementBlockType::Scope) {
-                return block;
-            }
-        }
-        auto *parent = stmt->parent();
-        if (parent->ir_node_kind() != IRNodeKind::StmtKind) {
-            throw InternalException("Non block statement's parent has to be a block statement");
-        }
-        auto *parent_stmt = reinterpret_cast<Stmt *>(parent);
-        return get_root_block(parent_stmt);
-    }
-
-    const std::string var_name_ = break_point_func_arg;
-    const uint32_t var_size_ = 32;
-
     Context *context_;
 };
 
@@ -136,6 +99,7 @@ public:
     void visit(FunctionCallStmt *stmt) override { extract_stmt_id(stmt); }
     void visit(ReturnStmt *stmt) override { extract_stmt_id(stmt); }
     void visit(AssertBase *stmt) override { extract_stmt_id(stmt); }
+    void visit(AuxiliaryStmt *stmt) override { extract_stmt_id(stmt); }
 
     const std::map<Stmt *, uint32_t> &map() const { return map_; }
 
@@ -273,6 +237,15 @@ void mock_hierarchy(Generator *top, const std::string &top_name) {
     }
 }
 
+void DebugDatabase::compute_generators(Generator *top) {
+    top_ = top;
+    GeneratorGraph g(top);
+    auto generators = g.get_sorted_nodes();
+    for (auto *gen : generators) {
+        generators_.emplace(gen);
+    }
+}
+
 void DebugDatabase::set_break_points(Generator *top) { set_break_points(top, ".py"); }
 
 void DebugDatabase::set_break_points(Generator *top, const std::string &ext) {
@@ -294,8 +267,8 @@ void DebugDatabase::set_break_points(Generator *top, const std::string &ext) {
             }
         }
     }
-    // set context
-    context_ = top->context();
+    // set generators
+    compute_generators(top);
 }
 
 void DebugDatabase::set_variable_mapping(
@@ -326,15 +299,6 @@ void DebugDatabase::set_variable_mapping(
     }
 }
 
-void DebugDatabase::set_generator_variable(
-    const std::map<Generator *, std::map<std::string, std::string>> &values) {
-    for (auto const &[gen, map] : values) {
-        for (auto const &[name, value] : map) {
-            generator_values_[gen].emplace(name, value);
-        }
-    }
-}
-
 class StmtContextVisitor : public IRVisitor {
 public:
     void visit(IfStmt *stmt) override { add_content(stmt); }
@@ -361,96 +325,93 @@ void DebugDatabase::set_stmt_context(kratos::Generator *top) {
     stmt_context_ = visitor.stmt_context();
 }
 
-class ConnectionVisitor : public IRVisitor {
+std::string get_trigger_condition(const Stmt *stmt) {
+    for (auto const &attr : stmt->get_attributes()) {
+        if (attr->type_str == "ssa-trigger") {
+            return attr->value_str;
+        }
+    }
+    return "";
+}
+
+std::optional<std::pair<std::string, std::string>> get_target_var_name(const Var *var) {
+    auto const &attrs = var->get_attributes();
+    for (auto const &attr : attrs) {
+        auto const &value_str = attr->value_str;
+        if (value_str.rfind("ssa=") == 0) {
+            auto pos = value_str.rfind(':');
+            auto scope_name = value_str.substr(4, pos - 4);
+            auto var_name = value_str.substr(pos + 1);
+            return std::make_pair(scope_name, var_name);
+        }
+    }
+    return std::nullopt;
+}
+
+void save_events(hgdb::DebugDatabase &db, Generator *top);
+
+class AssignVisitor : public IRVisitor {
 public:
-    void visit(Generator *generator) override {
-        lock_.lock();
-        generators_.emplace(generator);
-        lock_.unlock();
-        // loop through the module instance statement, where it holds the connection
-        // information
-        uint64_t child_count = generator->stmts_count();
-        for (uint64_t i = 0; i < child_count; i++) {
-            auto const stmt = generator->get_stmt(i);
-            if (stmt->type() == StatementType::ModuleInstantiation) {
-                auto mod = stmt->as<ModuleInstantiationStmt>();
-                const auto *target = mod->target();
-                auto target_handle_name = target->handle_name();
-                auto mapping = mod->port_mapping();
-                for (auto [target_port, parent_var] : mapping) {
-                    // we ignore the constant connection
-                    if (parent_var->type() != VarType::PortIO) {
-                        if (target_port->port_direction() == PortDirection::In) {
-                            auto source = parent_var->sources();
-                            if (source.size() == 1) {
-                                parent_var = (*source.begin())->right();
-                            }
-                        } else {
-                            auto sink = parent_var->sinks();
-                            if (sink.size() == 1) {
-                                parent_var = (*sink.begin())->left();
-                            }
-                        }
-                    }
-                    if (parent_var->type() != VarType::PortIO) {
-                        continue;
-                    }
-                    auto *gen = parent_var->generator();
-                    auto gen_handle = gen->handle_name();
-                    // the direction is var -> var
-                    lock_.lock();
-                    if (target_port->port_direction() == PortDirection::In) {
-                        connections_.emplace(
-                            std::make_pair(gen_handle, parent_var->to_string()),
-                            std::make_pair(target_handle_name, target_port->to_string()));
-                    } else {
-                        connections_.emplace(
-                            std::make_pair(target_handle_name, target_port->to_string()),
-                            std::make_pair(gen_handle, parent_var->to_string()));
-                    }
-                    lock_.unlock();
-                }
+    AssignVisitor(hgdb::DebugDatabase &db,
+                  const std::map<Stmt *, std::pair<std::string, uint32_t>> &breakpoints)
+        : db_(db), breakpoints_(breakpoints) {}
+
+    void visit(AssignStmt *stmt) override {
+        // some assignments may not be assigned properly
+        if (breakpoints_.find(stmt) == breakpoints_.end()) return;
+        // need to figure out the left variables
+        // notice that if the left is sliced by variable, we need to populate multiple
+        // additional conditions
+        auto const &left = stmt->left();
+        auto stmt_id = stmt->stmt_id();
+        bool sliced_by_var = false;
+
+        if (left->type() == VarType::Slice) {
+            auto const &slice = left->as<VarSlice>();
+            if (slice->sliced_by_var()) {
+                sliced_by_var = true;
+            }
+        }
+
+        if (sliced_by_var) {
+            // need to create duplicated statement with different condition based on the
+            // select var
+            auto const &var_var_slice = left->as<VarVarSlice>();
+            // for now, we only support one-level of slicing
+            auto const &size = var_var_slice->parent_var->size().front();
+            auto select_name = var_var_slice->sliced_var()->to_string();
+            auto base_name = var_var_slice->parent_var->to_string();
+            for (auto i = 0u; i < size; i++) {
+                auto transformed_name = fmt::format("{0}[{1}]", base_name, i);
+                auto cond = fmt::format("{0} == {1}", select_name, i);
+                // the usage is setting var[10] as a watch point
+                hgdb::store_assignment(db_, transformed_name, transformed_name, stmt_id, cond);
+            }
+        } else {
+            // need to compute the SSA transform to figure out the original variable mapping
+            // if any
+            auto mapping = get_target_var_name(left);
+            if (mapping) {
+                auto const &[var_name, _] = *mapping;
+                auto transformed_name = left->to_string();
+                hgdb::store_assignment(db_, var_name, transformed_name, stmt_id);
+            } else {
+                // no SSA, just use the default name for mapping
+                auto name = left->to_string();
+                hgdb::store_assignment(db_, name, name, stmt_id);
             }
         }
     }
 
-    const DebugDatabase::ConnectionMap &connections() const { return connections_; }
-    const std::unordered_set<Generator *> &generators() const { return generators_; }
-
 private:
-    DebugDatabase::ConnectionMap connections_;
-    std::unordered_set<Generator *> generators_;
-    std::mutex lock_;
+    hgdb::DebugDatabase &db_;
+    const std::map<Stmt *, std::pair<std::string, uint32_t>> &breakpoints_;
 };
 
-void DebugDatabase::set_generator_connection(kratos::Generator *top) {
-    // we create a IR visitor to visit the generators
-    ConnectionVisitor visitor;
-    // this can be parallelized
-    visitor.visit_generator_root_p(top);
-    connection_map_ = visitor.connections();
-    generators_ = visitor.generators();
-}
-
-class HierarchyVisitor : public IRVisitor {
-public:
-    void visit(Generator *generator) override {
-        auto handle_name = generator->handle_name();
-        for (auto const &gen : generator->get_child_generators()) {
-            hierarchy_.emplace_back(std::make_pair(handle_name, gen.get()));
-        }
-    }
-    const std::vector<std::pair<std::string, Generator *>> &hierarchy() const { return hierarchy_; }
-
-private:
-    std::vector<std::pair<std::string, Generator *>> hierarchy_;
-};
-
-void DebugDatabase::set_generator_hierarchy(kratos::Generator *top) {
-    HierarchyVisitor visitor;
-    // no parallel to make it in order
-    visitor.visit_generator_root(top);
-    hierarchy_ = visitor.hierarchy();
+void save_assignment(hgdb::DebugDatabase &db, Generator *top,
+                     const std::map<Stmt *, std::pair<std::string, uint32_t>> &breakpoints) {
+    AssignVisitor v(db, breakpoints);
+    v.visit_root(top);
 }
 
 void DebugDatabase::save_database(const std::string &filename, bool override) {
@@ -459,17 +420,14 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
             fs::remove(filename);
         }
     }
-    auto storage = init_storage(filename);
-    storage.sync_schema();
+    auto storage = hgdb::init_debug_db(filename);
+
+    // compute breakpoint conditions
+    auto breakpoint_conditions = compute_enable_condition(top_);
     // insert tables
-    // use transaction to speed up
-    auto guard = storage.transaction_guard();
-    // metadata
-    MetaData top_name{"top_name", top_name_};
-    storage.insert(top_name);
 
     // insert generator ids
-    std::unordered_map<Generator *, uint32_t> gen_id_map;
+    std::unordered_map<const Generator *, uint32_t> gen_id_map;
     std::unordered_map<std::string, uint32_t> handle_id_map;
     for (auto const &gen : generators_) {
         if (gen->generator_id < 0) {
@@ -479,38 +437,45 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
         int id = gen->generator_id;
         gen_id_map.emplace(gen, id);
         handle_id_map.emplace(gen->handle_name(), id);
-        Instance inst{id, gen->handle_name()};
-        storage.replace(inst);
+        hgdb::store_instance(storage, gen->generator_id, gen->handle_name());
     }
 
     // break points
     for (auto const &[stmt, id] : break_points_) {
         if (stmt_mapping_.find(stmt) != stmt_mapping_.end()) {
             auto const &[fn, ln] = stmt_mapping_.at(stmt);
-            BreakPoint br{id, fn, ln};
-            storage.replace(br);
+            auto const *gen = stmt->generator_parent();
+            auto instance_id = gen_id_map.at(gen);
+            std::string condition;
+            if (breakpoint_conditions.find(stmt) != breakpoint_conditions.end())
+                condition = breakpoint_conditions.at(stmt);
+            auto trigger_str = get_trigger_condition(stmt);
+            // we don't support column breakpoint since there is normally no such usage in
+            // Python
+            hgdb::store_breakpoint(storage, id, instance_id, fn, ln, 0, condition, trigger_str);
         }
     }
 
     // variables
     int variable_count = 0;
     std::unordered_set<Var *> var_id_set;
+    std::map<std::tuple<const int, std::string, std::string>, int> var_id_mapping;
+    std::unordered_map<Generator *, std::map<std::string, std::string>> self_context_mapping;
     // function to create variable and flatten the hierarchy
-    auto create_variable = [&](Var *var_, const int handle_id_, const std::string& name_,
-                               const std::string &value_, bool is_context_,
-                               uint32_t breakpoint_id_ = 0) {
-        Variable v;
-        v.is_var = var_ != nullptr;
-        v.is_context = is_context_;
-        v.handle = std::make_unique<int>(handle_id_);
-        v.name = "";
+    // NOLINTNEXTLINE
+    auto create_variable = [&](Var *var_, const int handle_id_, const std::string &name_,
+                               const std::string &value_, bool is_context_, uint32_t breakpoint_id_,
+                               bool gen_var = false) {
+        hgdb::Variable v;
+        v.is_rtl = var_ != nullptr;
 
-        auto add_context = [&]() {
+        auto add_context_gen = [&](const std::string &name) {
             if (is_context_) {
                 // create context mapping as well
-                ContextVariable c_v{std::make_unique<uint32_t>(breakpoint_id_),
-                                    std::make_unique<int>(v.id), v.name};
-                storage.replace(c_v);
+                hgdb::store_context_variable(storage, name, breakpoint_id_, v.id);
+            }
+            if (gen_var) {
+                hgdb::store_generator_variable(storage, name, handle_id_, v.id);
             }
         };
 
@@ -520,55 +485,78 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
                 // it's an array. need to flatten it
                 auto slices = get_flatten_slices(var_);
                 for (auto const &slice : slices) {
-                    std::string new_name;
-                    if (!name_.empty()) {
-                        new_name = name_;
-                        for (auto const &s: slice)
-                            new_name = ::format("{0}.{1}", new_name, s);
-                    }
+                    std::string new_name = name_;
+                    for (auto const &s : slice) new_name = ::format("{0}.{1}", new_name, s);
                     std::string value = var_->name;
-                    for (auto const &s: slice)
-                        value = ::format("{0}[{1}]", value, s);
-                    v.name = new_name;
+                    for (auto const &s : slice) value = ::format("{0}[{1}]", value, s);
                     v.value = value;
-                    v.id = variable_count++;
-                    storage.replace(v);
-                    add_context();
+                    if (var_id_mapping.find(std::make_tuple(handle_id_, new_name, v.value)) ==
+                        var_id_mapping.end()) {
+                        v.id = variable_count++;
+                        hgdb::store_variable(storage, v.id, v.value);
+                        var_id_mapping.emplace(
+                            std::make_pair(std::make_tuple(handle_id_, new_name, v.value), v.id));
+                    } else {
+                        v.id = var_id_mapping.at(std::make_tuple(handle_id_, new_name, v.value));
+                    }
+                    add_context_gen(new_name);
                 }
             } else if (var_->is_struct()) {
-                // it's an packed array
+                // it's a packed array
                 if (var_->type() == VarType::PortIO) {
                     auto *p = reinterpret_cast<PortPackedStruct *>(var_);
                     auto const &def = p->packed_struct();
-                    for (auto const &iter : def.attributes) {
-                        auto const &attr_name = std::get<0>(iter);
+                    for (auto const &iter : def->attributes) {
+                        auto const &attr_name = iter.name;
                         // we need to store lots of them
-                        if (!name_.empty()) v.name = ::format("{0}.{1}", name_, attr_name);
+                        std::string new_name = ::format("{0}.{1}", name_, attr_name);
                         v.value = ::format("{0}.{1}", var_->name, attr_name);
-                        v.id = variable_count++;
-                        storage.replace(v);
-                        add_context();
+                        if (var_id_mapping.find(std::make_tuple(handle_id_, new_name, v.value)) ==
+                            var_id_mapping.end()) {
+                            v.id = variable_count++;
+                            hgdb::store_variable(storage, v.id, v.value);
+                            var_id_mapping.emplace(std::make_pair(
+                                std::make_tuple(handle_id_, new_name, v.value), v.id));
+                        } else {
+                            v.id =
+                                var_id_mapping.at(std::make_tuple(handle_id_, new_name, v.value));
+                        }
+                        add_context_gen(new_name);
                     }
                 } else if (var_->type() == VarType::Base) {
                     auto *p = reinterpret_cast<VarPackedStruct *>(var_);
                     auto const &def = p->packed_struct();
-                    for (auto const &iter : def.attributes) {
-                        auto const &attr_name = std::get<0>(iter);
+                    for (auto const &iter : def->attributes) {
+                        auto const &attr_name = iter.name;
                         // we need to store lots of them
-                        if (!name_.empty()) v.name = ::format("{0}.{1}", name_, attr_name);
+                        std::string new_name = ::format("{0}.{1}", name_, attr_name);
                         v.value = ::format("{0}.{1}", var_->name, attr_name);
-                        v.id = variable_count++;
-                        storage.replace(v);
-                        add_context();
+                        if (var_id_mapping.find(std::make_tuple(handle_id_, new_name, v.value)) ==
+                            var_id_mapping.end()) {
+                            v.id = variable_count++;
+                            hgdb::store_variable(storage, v.id, v.value);
+                            var_id_mapping.emplace(std::make_pair(
+                                std::make_tuple(handle_id_, new_name, v.value), v.id));
+                        } else {
+                            v.id =
+                                var_id_mapping.at(std::make_tuple(handle_id_, new_name, v.value));
+                        }
+                        add_context_gen(new_name);
                     }
                 }
             } else {
                 // the normal one
-                v.name = name_;
                 v.value = var_->name;
-                v.id = variable_count++;
-                storage.replace(v);
-                add_context();
+                if (var_id_mapping.find(std::make_tuple(handle_id_, name_, v.value)) ==
+                    var_id_mapping.end()) {
+                    v.id = variable_count++;
+                    hgdb::store_variable(storage, v.id, v.value);
+                    var_id_mapping.emplace(
+                        std::make_pair(std::make_tuple(handle_id_, name_, v.value), v.id));
+                } else {
+                    v.id = var_id_mapping.at(std::make_tuple(handle_id_, name_, v.value));
+                }
+                add_context_gen(name_);
             }
             var_id_set.emplace(var_);
         } else {
@@ -576,71 +564,27 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
             if (name_.empty()) {
                 throw UserException(::format("Non-variable cannot have empty name in database"));
             }
-            v.name = name_;
             v.value = value_;
             v.id = variable_count++;
-            storage.replace(v);
-            add_context();
+            hgdb::store_variable(storage, v.id, v.value, false);
+            add_context_gen(name_);
         }
     };
     for (auto const &[handle_name, gen_map] : variable_mapping_) {
         auto const &[gen, vars] = gen_map;
         if (gen_id_map.find(gen) == gen_id_map.end())
             throw InternalException(::format("Unable to find generator {0}", gen->handle_name()));
+        self_context_mapping.emplace(gen, vars);
         auto id = gen_id_map.at(gen);
-        for (auto const &[front_var, var] : vars) {
-            auto gen_var = gen->get_var(var);
-            if (!gen_var) throw InternalException(::format("Unable to get variable {0}", var));
-            create_variable(gen_var.get(), id, front_var, "", false);
-        }
+        // this will be generator instance values (RTL correspondence)
         auto all_vars = gen->get_all_var_names();
         for (auto const &var_name : all_vars) {
             auto var = gen->get_var(var_name);
-            // continue because we already have that variable stored
-            if (var_id_set.find(var.get()) != var_id_set.end()) continue;
             if (var && (var->type() == VarType::Base || var->type() == VarType::PortIO)) {
-                create_variable(var.get(), id, "", "", false);
+                create_variable(var.get(), id, var_name, var_name, false, 0, true);
             }
         }
     }
-    for (auto const &[gen, map] : generator_values_) {
-        auto const handle_name = gen->handle_name();
-        if (handle_id_map.find(handle_name) == handle_id_map.end())
-            throw InternalException(::format("Unable to find id for {0}", handle_name));
-        auto id = handle_id_map.at(handle_name);
-        for (auto const &[name, value] : map) {
-            auto var = gen->get_var(name);
-            create_variable(var.get(), id, name, value, false);
-        }
-    }
-
-    // connections
-    for (auto const &[from, to] : connection_map_) {
-        auto const [from_handle, from_var] = from;
-        auto const [to_handle, to_var] = to;
-        if (handle_id_map.find(from_handle) == handle_id_map.end())
-            throw InternalException(::format("Unable to find id for {0}", from_handle));
-        if (handle_id_map.find(to_handle) == handle_id_map.end())
-            throw InternalException(::format("Unable to find id for {0}", to_handle));
-        auto from_id = handle_id_map.at(from_handle);
-        auto to_id = handle_id_map.at(to_handle);
-        Connection conn{std::make_unique<int>(from_id), from_var, std::make_unique<int>(to_id),
-                        to_var};
-        storage.replace(conn);
-    }
-
-    // hierarchy
-    for (auto const &[handle, child] : hierarchy_) {
-        if (handle_id_map.find(handle) == handle_id_map.end())
-            throw InternalException(::format("Unable to find id for {0}", handle));
-        if (handle_id_map.find(child->handle_name()) == handle_id_map.end())
-            throw InternalException(::format("Unable to find id for {0}", child->handle_name()));
-        auto id = handle_id_map.at(handle);
-        auto child_id = handle_id_map.at(child->handle_name());
-        Hierarchy h{std::make_unique<int>(id), child->instance_name,
-                    std::make_unique<int>(child_id)};
-        storage.replace(h);
-    }
 
     // local context variables
     for (auto const &[stmt, id] : break_points_) {
@@ -648,9 +592,11 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
         auto const &gen = stmt->generator_parent();
         auto handle_name = gen->handle_name();
         if (stmt_context_.find(stmt) == stmt_context_.end()) continue;
+        if (stmt_mapping_.find(stmt) == stmt_mapping_.end()) continue;
         if (handle_id_map.find(handle_name) == handle_id_map.end())
             throw InternalException(::format("Unable to find {0} in handle id map", handle_name));
 
+        auto const &self_vars = self_context_mapping.at(gen);
         auto instance_id = handle_id_map.at(handle_name);
         auto values = stmt_context_.at(stmt);
 
@@ -659,16 +605,24 @@ void DebugDatabase::save_database(const std::string &filename, bool override) {
             auto *gen_var = is_var ? gen->get_var(value).get() : nullptr;
             create_variable(gen_var, instance_id, key, value, true, id);
         }
+
+        // self/this instance
+        for (auto const &[front_var, var] : self_vars) {
+            Var *gen_var = gen->get_var(var).get();
+            if (!gen_var) {
+                gen_var = gen->get_param(var).get();
+            }
+            if (!gen_var) {
+                create_variable(nullptr, instance_id, front_var, var, true, id);
+            } else {
+                create_variable(gen_var, instance_id, front_var, var, true, id);
+            }
+        }
     }
 
-    // instance id set
-    for (auto const &[stmt, id] : break_points_) {
-        auto gen_id = stmt->generator_parent()->generator_id;
-        InstanceSetEntry entry{std::make_unique<int>(gen_id), std::make_unique<int>(id)};
-        storage.replace(entry);
-    }
+    save_events(storage, top_);
 
-    guard.commit();
+    save_assignment(storage, top_, stmt_mapping_);
 }
 
 void inject_clock_break_points(Generator *top) {
