@@ -25,8 +25,9 @@ using std::vector;
 namespace kratos {
 
 bool is_relational_op(ExprOp op) {
-    static std::unordered_set<ExprOp> ops = {ExprOp::LessThan, ExprOp::GreaterThan,
-                                             ExprOp::LessEqThan, ExprOp::GreaterEqThan, ExprOp::Eq};
+    static std::unordered_set<ExprOp> ops = {ExprOp::LessThan,   ExprOp::GreaterThan,
+                                             ExprOp::LessEqThan, ExprOp::GreaterEqThan,
+                                             ExprOp::Eq,         ExprOp::Neq};
     return ops.find(op) != ops.end();
 }
 
@@ -40,6 +41,8 @@ bool is_expand_op(ExprOp op) {
     return ops.find(op) != ops.end();
 }
 
+bool is_ternary_op(ExprOp op) { return op == ExprOp::Conditional; }
+
 bool is_unary_op(ExprOp op) {
     static std::unordered_set<ExprOp> ops = {ExprOp::UXor,  ExprOp::UPlus, ExprOp::UOr,
                                              ExprOp::UNot,  ExprOp::UXor,  ExprOp::UInvert,
@@ -192,14 +195,25 @@ VarSlice &Var::operator[](std::pair<uint32_t, uint32_t> slice) {
     // create a new one
     // notice that slice is not part of generator's variables. It's handled by the parent (var)
     // itself
+    std::shared_ptr<VarSlice> var_slice = ::make_shared<VarSlice>(this, high, low);
+    if (width_param_) {
+        var_slice->set_width_param(width_param_);
+    }
     // depends on the root variable, if it is actually a struct, we need to return to the
     // actual trampoline class as proxy
-    auto var_root = get_var_root_parent();
-    std::shared_ptr<VarSlice> var_slice = ::make_shared<VarSlice>(this, high, low);
-    if (var_root->is_struct() && var_slice->width() == var_slice->var_width()) {
+    if (get_var_root_parent()->is_struct() && var_slice->width() == var_slice->var_width()) {
         // we actually reached the real struct
+        auto *s = dynamic_cast<PackedInterface *>(get_var_root_parent());
+        auto packed_slice = std::make_shared<PackedSlice>(var_slice.get(), var_slice->width() - 1,
+                                                          0, s->packed_struct().get());
+        slices_.emplace_back(packed_slice);
+        // this is a hack here, but we store the parent into the child
+        packed_slice->slices_.emplace_back(var_slice);
+        var_slice = packed_slice;
+    } else {
+        slices_.emplace_back(var_slice);
     }
-    slices_.emplace_back(var_slice);
+
     return *var_slice;
 }
 
@@ -213,6 +227,9 @@ VarSlice &Var::operator[](const std::shared_ptr<Var> &var) {
     }
     auto var_slice = ::make_shared<VarVarSlice>(this, var.get());
     slices_.emplace_back(var_slice);
+    if (width_param_) {
+        var_slice->set_width_param(width_param_);
+    }
     return *var_slice;
 }
 
@@ -282,7 +299,6 @@ std::string Var::handle_name(kratos::Generator *scope) const {
 
 void Var::set_width_param(const std::shared_ptr<Var> &param) { set_width_param(param.get()); }
 
-
 class ParamVisitor : public IRVisitor {
 public:
     void visit(Param *param) override { params_.emplace(param); }
@@ -309,7 +325,6 @@ void Var::set_width_param(Var *param) {
 
 VarSlice &Var::operator[](uint32_t bit) { return this->operator[]({bit, bit}); }
 
-
 void Var::set_size_param(uint32_t index, Var *param) {
     if (type_ != VarType::Base && type_ != VarType::PortIO)
         throw UserException(::format("{0} is not a variable or port", to_string()));
@@ -465,16 +480,16 @@ VarSlice::VarSlice(Var *parent, uint32_t high, uint32_t low)
         // the wire, which means we need to count the number of size decreases
         // this is the actual slice
         var_width_ = high - low + 1;
-        is_packed_ = false;
+        is_packed_ = parent->is_packed();
     } else if (parent->size().size() == 1 && parent->size().front() == 1 && dropped_dim_size1) {
         // need to keep the var width calculation correct
         var_width_ = parent->var_width();
-        is_packed_ = false;
+        is_packed_ = true;
     } else {
         if (high == low) {
             if (parent->size().size() == 1) {
                 size_ = {1};
-                is_packed_ = false;
+                is_packed_ = true;
             } else {
                 size_ = std::vector<uint32_t>(parent->size().begin() + 1, parent->size().end());
                 is_packed_ = parent->is_packed();
@@ -557,12 +572,27 @@ std::vector<std::pair<uint32_t, uint32_t>> VarSlice::get_slice_index() const {
 }
 
 PackedSlice &VarSlice::operator[](const std::string &member_name) {
-    auto const *root = get_var_root_parent();
-    if (!root->is_packed() || width() != var_width())
+    if (!parent_var) throw InternalException("Invalid var slice");
+    if (!parent_var->is_packed() || width() != var_width())
         throw UserException(::format("Unable to access {0}.{1}", to_string(), member_name));
-    auto p = std::make_shared<PackedSlice>(this, true);
+    auto *root = get_var_root_parent();
+    if (!root->is_struct())
+        throw UserException(::format("Invalid member access{0}.{1}", to_string(), member_name));
+    const PackedStructFieldDef *def;
+    if (is_struct()) {
+        auto *packed = dynamic_cast<PackedSlice *>(this);
+        return packed->slice_member(member_name);
+    } else {
+        auto *packed = dynamic_cast<PackedInterface *>(root);
+        if (!packed)
+            throw UserException(::format("Unable to access {0}.{1}", to_string(), member_name));
+        def = packed->get_definition(member_name);
+    }
+
+    if (!def) throw UserException(::format("Unable to access {0}.{1}", to_string(), member_name));
+    auto p = std::make_shared<PackedSlice>(this, def, member_name);
     slices_.emplace_back(p);
-    return p->slice_member(member_name);
+    return *p;
 }
 
 VarVarSlice::VarVarSlice(kratos::Var *parent, kratos::Var *slice)
@@ -655,7 +685,8 @@ Expr::Expr(ExprOp op, Var *left, Var *right)
     if (right != nullptr && left->width() != right->width()) {
         // see if we can resize
         if (IterVar::safe_to_resize(left, right->width(), right->is_signed()) &&
-            (right->type() != VarType::ConstValue && right->type() != VarType::Parameter)) {
+            (right->type() != VarType::ConstValue && right->type() != VarType::Parameter &&
+             right->type() != VarType::Iter)) {
             // this is a hack
             if (left->type() == VarType::ConstValue) {
                 left->var_width() = right->width();
@@ -690,7 +721,7 @@ Expr::Expr(ExprOp op, Var *left, Var *right)
         var_width_ = left->var_width();
 
     if (right != nullptr)
-        is_signed_ = left->is_signed() & right->is_signed();
+        is_signed_ = left->is_signed() && right->is_signed();
     else
         is_signed_ = left->is_signed();
     type_ = VarType::Expression;
@@ -803,14 +834,33 @@ std::shared_ptr<AssignStmt> Var::assign(const std::shared_ptr<Var> &var, Assignm
     return assign_(var, type);
 }
 
+class AllConcat : public IRVisitor {
+public:
+    void visit(Expr *expr) override {
+        if (expr->op != ExprOp::Concat) {
+            legal = false;
+        }
+    }
+    bool legal = true;
+};
+
 std::shared_ptr<AssignStmt> Var::assign_(const std::shared_ptr<Var> &var, AssignmentType type) {
+    if (!var) {
+        throw UserException("Trying to assign null to " + to_string());
+    }
     // if it's a constant or expression, it can't be assigned to
-    if (type_ == VarType::ConstValue)
+    if (type_ == VarType::ConstValue) {
         throw VarException(::format("Cannot assign {0} to a const {1}", var->to_string(), name),
                            {this, var.get()});
-    else if (type_ == VarType::Expression)
-        throw VarException(::format("Cannot assign {0} to an expression", var->to_string(), name),
-                           {this, var.get()});
+    } else if (type_ == VarType::Expression) {
+        auto all_concat = AllConcat();
+        all_concat.visit_root(this);
+        if (!all_concat.legal)
+            throw VarException(
+                ::format("Cannot assign {0} to an expression", var->to_string(), name),
+                {this, var.get()});
+    }
+
     auto stmt = ::make_shared<AssignStmt>(shared_from_this(), var, type);
 
     return stmt;
@@ -835,7 +885,7 @@ Const::Const(Generator *generator, int64_t value, uint32_t width, bool is_signed
         std::memcpy(&min, &temp, sizeof(min));
         throw UserException(::format("{0} is smaller than the minimum value ({1}) given width {2}",
                                      value, min, width));
-    } else if (is_legal == ConstantLegal::Big) {
+    } else if (is_legal == ConstantLegal::Big && width <= 64) {
         uint64_t max = (1ull << width) - 1;
         uint64_t unsigned_value;
         std::memcpy(&unsigned_value, &value, sizeof(unsigned_value));
@@ -857,7 +907,7 @@ Const::ConstantLegal Const::is_legal(int64_t value, uint32_t width, bool is_sign
         std::memcpy(&max, &temp, sizeof(max));
         if (value > max) return Const::ConstantLegal::Big;
     } else {
-        uint64_t max = (1ull << width) - 1;
+        uint64_t max = width == 64 ? std::numeric_limits<uint64_t>::max() : (1ull << width) - 1;
         uint64_t unsigned_value;
         std::memcpy(&unsigned_value, &value, sizeof(unsigned_value));
         if (unsigned_value > max) return Const::ConstantLegal::Big;
@@ -877,6 +927,19 @@ Const &Const::constant(int64_t value, uint32_t width, bool is_signed) {
     return *p;
 }
 
+Const &Const::constant(const std::string &hex_value, uint32_t num_bits, bool negative,
+                       uint32_t width, bool is_signed) {
+    auto p = std::make_shared<Const>(std::numeric_limits<uint64_t>::max(), width, is_signed);
+    if (hex_value.find_first_not_of("abcdefABCDEF0123456789") != std::string::npos) {
+        throw std::invalid_argument("Invalid hex value for constant " + hex_value);
+    }
+    p->hex_value_ = hex_value;
+    p->negative_ = negative;
+    p->num_bits_ = num_bits;
+    consts_.emplace(p);
+    return *p;
+}
+
 void Const::set_is_packed(bool value) {
     if (!value) throw UserException("Unable to set const unpacked");
 }
@@ -940,6 +1003,9 @@ void VarCasted::set_target_width(uint32_t width) {
 }
 
 void VarCasted::add_sink(const std::shared_ptr<AssignStmt> &stmt) { parent_var_->add_sink(stmt); }
+void VarCasted::add_source(const std::shared_ptr<AssignStmt> &stmt) {
+    parent_var_->add_source(stmt);
+}
 
 std::shared_ptr<Var> Var::cast(VarCastType cast_type) {
     if (cast_type == VarCastType::Signed && is_signed_) {
@@ -979,14 +1045,14 @@ void Const::add_source(const std::shared_ptr<AssignStmt> &) {
 }
 
 void Const::add_sink(const std::shared_ptr<AssignStmt> &stmt) {
-    auto left = stmt->left();
+    auto *left = stmt->left();
     // if it's a port, we change the constant's generator to that of port
-    auto generator = left->generator();
+    auto *generator = left->generator();
     if (!generator)
         throw StmtException(::format("Unable to find left hand side generator"), {stmt.get()});
-    auto parent = generator->parent();
+    auto *parent = generator->parent();
     if (parent && parent->ir_node_kind() == GeneratorKind) {
-        auto gen = dynamic_cast<Generator *>(parent);
+        auto *gen = dynamic_cast<Generator *>(parent);
         this->generator_ = gen;
     }
 }
@@ -1160,7 +1226,7 @@ VarExtend::VarExtend(const std::shared_ptr<Var> &var, uint32_t width)
     var_width_ = width;
     is_signed_ = parent_->is_signed();
     if (parent_->size().size() > 1 || parent_->size().front() > 1 ||
-        (parent_->is_packed() && parent_->type() != VarType::ConstValue)) {
+        (parent_->explicit_array() && parent_->type() != VarType::ConstValue)) {
         throw VarException(::format("Cannot extend an array ({0})", parent_->to_string()),
                            {parent_});
     }
@@ -1185,7 +1251,14 @@ std::string VarExtend::to_string() const {
 }
 
 std::string Const::to_string() const {
-    if (is_signed_ && value_ < 0) {
+    if (num_bits_ != 0) {
+        // very big number
+        if (negative_) {
+            return ::format("-{0}\'h{1}", width(), hex_value_);
+        } else {
+            return ::format("{0}\'h{1}", width(), hex_value_);
+        }
+    } else if (is_signed_ && value_ < 0) {
         return ::format("-{0}\'h{1:X}", width(), -value_);
     } else {
         return ::format("{0}\'h{1:X}", width(), value_);
@@ -1462,7 +1535,7 @@ void Var::move_linked_to(kratos::Var *new_var) {
 
     // change concat'ed vars
     // we use overloaded ones
-    for (auto &concat : concat_vars_) {
+    for (auto const &concat : concat_vars_) {
         concat->move_linked_to(new_var);
         concat->replace_var(shared_from_this(), new_var->shared_from_this());
     }
@@ -1471,7 +1544,7 @@ void Var::move_linked_to(kratos::Var *new_var) {
     concat_vars_.clear();
 
     // casted
-    for (auto &var : casted_) {
+    for (auto const &var : casted_) {
         var->set_parent(new_var->get_var_root_parent());
         new_var->get_var_root_parent()->casted_.emplace(var);
     }
@@ -1480,7 +1553,7 @@ void Var::move_linked_to(kratos::Var *new_var) {
 
 void Var::clear_sources(bool remove_parent) {  // NOLINT
     if (remove_parent) {
-        for (auto &stmt : sources_) {
+        for (auto const &stmt : sources_) {
             stmt->remove_from_parent();
         }
     }
@@ -1490,7 +1563,7 @@ void Var::clear_sources(bool remove_parent) {  // NOLINT
 
 void Var::clear_sinks(bool remove_parent) {  // NOLINT
     if (remove_parent) {
-        for (auto &stmt : sinks_) {
+        for (auto const &stmt : sinks_) {
             if (stmt->parent()) stmt->remove_from_parent();
         }
     }
@@ -1566,49 +1639,104 @@ std::string ConditionalExpr::handle_name(kratos::Generator *scope) const {
                     right->handle_name(scope));
 }
 
+uint32_t PackedStructFieldDef::bitwidth() const {
+    if (struct_) {
+        return struct_->bitwidth();
+    } else {
+        return width;
+    }
+}
+
+bool PackedStructFieldDef::same(const PackedStructFieldDef &def) const {
+    if (struct_ && def.struct_) {
+        return struct_->same(*def.struct_);
+    } else if (!struct_ && !def.struct_) {
+        return name == def.name && width == def.width && signed_ == def.signed_;
+    }
+    return false;
+}
+
 PackedStruct::PackedStruct(std::string struct_name,
-                           std::vector<std::tuple<std::string, uint32_t, bool>> attributes)
-    : struct_name(std::move(struct_name)), attributes(std::move(attributes)) {}
+                           const std::vector<std::tuple<std::string, uint32_t, bool>> &attributes)
+    : struct_name(std::move(struct_name)) {
+    for (auto const &[name, size, signed_] : attributes) {
+        auto def = PackedStructFieldDef();
+        def.signed_ = signed_;
+        def.name = name;
+        def.width = size;
+        this->attributes.emplace_back(def);
+    }
+}
 
 PackedStruct::PackedStruct(std::string struct_name,
                            const std::vector<std::tuple<std::string, uint32_t>> &attributes)
     : struct_name(std::move(struct_name)) {
     for (auto const &[name, size] : attributes) {
-        this->attributes.emplace_back(std::make_tuple(name, size, false));
+        auto def = PackedStructFieldDef();
+        def.signed_ = false;
+        def.name = name;
+        def.width = size;
+        this->attributes.emplace_back(def);
     }
 }
 
+uint32_t PackedStruct::bitwidth() const {
+    uint32_t result = 0;
+    for (auto const &def : attributes) {
+        result += def.bitwidth();
+    }
+    return result;
+}
+
+bool PackedStruct::same(const PackedStruct &def) const {
+    if (attributes.size() != def.attributes.size()) return false;
+    for (auto i = 0u; i < attributes.size(); i++) {
+        if (!attributes[i].same(def.attributes[i])) return false;
+    }
+    return true;
+}
+
 PackedSlice::PackedSlice(PortPackedStruct *parent, const std::string &member_name)
-    : VarSlice(parent, 0, 0), member_name_(member_name) {
+    : VarSlice(parent, 0, 0) {
     auto const &struct_ = parent->packed_struct();
-    set_up(struct_, member_name);
+    set_up(*struct_, member_name);
 }
 
 PackedSlice::PackedSlice(kratos::VarPackedStruct *parent, const std::string &member_name)
-    : VarSlice(parent, 0, 0), member_name_(member_name) {
+    : VarSlice(parent, 0, 0) {
     auto const &struct_ = parent->packed_struct();
-    set_up(struct_, member_name);
+    set_up(*struct_, member_name);
 }
 
-PackedSlice::PackedSlice(VarSlice *slice, bool is_root)
-    : VarSlice(slice, 0, 0), is_root_(is_root) {}
+PackedSlice::PackedSlice(Var *parent, uint32_t high, uint32_t low, const PackedStruct *struct_)
+    : VarSlice(parent, high, low), original_struct_(struct_) {}
+
+PackedSlice::PackedSlice(Var *parent, const PackedStructFieldDef *def,
+                         const std::string &member_name)
+    : VarSlice(parent, 0, 0), def_(def) {
+    if (!def->struct_) {
+        throw InternalException("Invalid member access " + def->name);
+    }
+    set_up(*def->struct_, member_name);
+}
 
 void PackedSlice::set_up(const kratos::PackedStruct &struct_, const std::string &member_name) {
     // compute the high and low
     uint32_t low_ = 0;
     bool found = false;
-    for (auto const &[name, width, is_signed] : struct_.attributes) {
-        if (name == member_name) {
+    for (auto const &def : struct_.attributes) {
+        if (def.name == member_name) {
             found = true;
-            high = width + low_ - 1;
+            high = def.bitwidth() + low_ - 1;
             low = low_;
-            is_signed_ = is_signed;
-            var_high_ = high;
-            var_low_ = low;
+            is_signed_ = def.signed_;
+            var_high_ = high + parent_var->var_low();
+            var_low_ = low + parent_var->var_low();
             var_width_ = var_high_ - var_low_ + 1;
+            def_ = &def;
             break;
         } else {
-            low_ += width;
+            low_ += def.bitwidth();
         }
     }
 
@@ -1619,39 +1747,42 @@ void PackedSlice::set_up(const kratos::PackedStruct &struct_, const std::string
 }
 
 shared_ptr<Var> PackedSlice::slice_var(std::shared_ptr<Var> var) {
-    if (is_root_) return var;
     if (var->type() == VarType::PortIO) {
         auto v = var->as<PortPackedStruct>();
-        return v->operator[](member_name_).shared_from_this();
+        return v->operator[](def_->name).shared_from_this();
     } else {
         auto v = var->as<VarPackedStruct>();
-        return v->operator[](member_name_).shared_from_this();
+        return v->operator[](def_->name).shared_from_this();
     }
 }
 
 PackedSlice &PackedSlice::slice_member(const std::string &member_name) {
-    if (!is_root_) throw UserException("Only struct array can slice member at the end");
-    auto *root = get_var_root_parent();
-    std::shared_ptr<PackedSlice> p;
-    if (root->type() == VarType::PortIO) {
-        auto v = root->as<PortPackedStruct>();
-        p = ::make_shared<PackedSlice>(this, false);
-        p->set_up(v->packed_struct(), member_name);
+    const PackedStruct *struct_;
+    if (original_struct_) {
+        struct_ = original_struct_;
+    } else if (is_struct()) {
+        struct_ = def_->struct_;
     } else {
-        auto v = root->as<VarPackedStruct>();
-        p = ::make_shared<PackedSlice>(this, false);
-        p->set_up(v->packed_struct(), member_name);
+        auto *root = get_var_root_parent();
+        if (root->type() == VarType::PortIO) {
+            auto v = root->as<PortPackedStruct>();
+            struct_ = v->packed_struct().get();
+        } else {
+            auto v = root->as<VarPackedStruct>();
+            struct_ = v->packed_struct().get();
+        }
     }
-    p->member_name_ = member_name;
+    std::shared_ptr<PackedSlice> p = ::make_shared<PackedSlice>(this);
+    p->set_up(*struct_, member_name);
     slices_.emplace_back(p);
     return *p;
 }
 
 std::string PackedSlice::to_string() const {
-    if (is_root_) {
+    if (original_struct_) {
         return parent_var->to_string();
     } else {
-        return ::format("{0}.{1}", parent_var->to_string(), member_name_);
+        return ::format("{0}.{1}", parent_var->to_string(), def_->name);
     }
 }
 
@@ -1664,18 +1795,20 @@ PackedSlice &VarPackedStruct::operator[](const std::string &member_name) {
     return *ptr;
 }
 
-VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name, PackedStruct packed_struct_)
+VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name,
+                                 std::shared_ptr<PackedStruct> packed_struct_)
     : Var(m, name, 1, 1, false), struct_(std::move(packed_struct_)) {
     compute_width();
 }
 
-VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name, PackedStruct packed_struct_,
-                                 uint32_t size)
+VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name,
+                                 std::shared_ptr<PackedStruct> packed_struct_, uint32_t size)
     : Var(m, name, 1, size, false), struct_(std::move(packed_struct_)) {
     compute_width();
 }
 
-VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name, PackedStruct packed_struct_,
+VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name,
+                                 std::shared_ptr<PackedStruct> packed_struct_,
                                  const std::vector<uint32_t> &size)
     : Var(m, name, 1, size, false), struct_(std::move(packed_struct_)) {
     compute_width();
@@ -1683,21 +1816,26 @@ VarPackedStruct::VarPackedStruct(Generator *m, const std::string &name, PackedSt
 
 void VarPackedStruct::compute_width() {
     // compute the width
-    uint32_t width = 0;
-    for (auto const &def : struct_.attributes) {
-        width += std::get<1>(def);
-    }
-    var_width_ = width;
+    var_width_ = struct_->bitwidth();
 }
 
 std::set<std::string> VarPackedStruct::member_names() const {
     std::set<std::string> result;
-    for (auto const &def : struct_.attributes) {
-        result.emplace(std::get<0>(def));
+    for (auto const &def : struct_->attributes) {
+        result.emplace(def.name);
     }
     return result;
 }
 
+const PackedStructFieldDef *VarPackedStruct::get_definition(const std::string &name) const {
+    for (auto const &def : struct_->attributes) {
+        if (name == def.name) {
+            return &def;
+        }
+    }
+    return nullptr;
+}
+
 void VarPackedStruct::set_is_packed(bool value) {
     if (!value) throw UserException("Unable to set packed struct unpacked");
 }
@@ -1782,7 +1920,7 @@ void extract_iter_var(const Var *var, std::vector<const IterVar *> &iters) {
     if (var->type() == VarType::Iter) {
         iters.emplace_back(reinterpret_cast<const IterVar *>(var));
     } else if (var->type() == VarType::Expression) {
-        auto expr = reinterpret_cast<const Expr *>(var);
+        auto const *expr = reinterpret_cast<const Expr *>(var);
         extract_iter_var(expr->left, iters);
         extract_iter_var(expr->right, iters);
     }
@@ -1793,9 +1931,9 @@ std::optional<int64_t> convert_value(const Var *var,
     if (!values) {
         return std::nullopt;
     }
-    auto v = &((*values)[0]);
+    auto const *v = &((*values)[0]);
     if (var->is_signed()) {
-        auto s_v = reinterpret_cast<const int64_t *>(v);
+        auto const *s_v = reinterpret_cast<const int64_t *>(v);
         return *s_v;
     } else {
         return *v;
@@ -1847,7 +1985,7 @@ bool IterVar::safe_to_resize(const Var *var, uint32_t target_size, bool is_signe
     // brute-force to compute every possible combinations using simulator
     static Simulator sim(nullptr);
     // We did some hacks that, assuming the min and max value is at the boundary
-    // maybe use a SMT solver?
+    // maybe use an SMT solver?
     std::queue<const IterVar *> queue;
     for (auto const &v : iters) {
         sim.set_i(v, v->max_value() - 1, false);
@@ -1863,10 +2001,10 @@ bool IterVar::has_iter_var(const Var *var) {
     if (!var) return false;
     if (var->type() == VarType::Iter) return true;
     if (var->type() == VarType::Slice) {
-        auto slice = reinterpret_cast<const VarSlice *>(var);
+        auto const *slice = reinterpret_cast<const VarSlice *>(var);
         return has_iter_var(slice->get_var_root_parent());
     } else if (var->type() == VarType::Expression) {
-        auto expr = reinterpret_cast<const Expr *>(var);
+        auto const *expr = reinterpret_cast<const Expr *>(var);
         return has_iter_var(expr->left) || has_iter_var(expr->right);
     }
     return false;
@@ -1906,7 +2044,7 @@ std::shared_ptr<AssignStmt> EnumVar::assign_(const std::shared_ptr<Var> &var,
         if (p->enum_def()->name != enum_type_->name)
             throw VarException("Cannot assign different enum type", {this, var.get()});
     } else {
-        auto p = dynamic_cast<EnumType *>(var.get());
+        auto *p = dynamic_cast<EnumType *>(var.get());
         if (!p) throw InternalException("Unable to obtain enum definition");
         if (!p->enum_type())
             throw VarException(::format("Cannot obtain enum information from var ({0}). "
@@ -1931,7 +2069,7 @@ FunctionCallVar::FunctionCallVar(Generator *m, const std::shared_ptr<FunctionStm
             throw VarException(::format("{0} is not connected", port_name), {func_port.get()});
         }
         // check the port types
-        auto &arg_port = args.at(port_name);
+        auto const &arg_port = args.at(port_name);
         if (func_port->width() != arg_port->width())
             throw VarException(::format("{0}'s width doesn't match", port_name),
                                {func_port.get(), arg_port.get()});
