@@ -64,7 +64,7 @@ std::vector<FSMState*> FSM::get_all_child_states(bool include_extra_state) const
         queue.emplace(iter.second.get());
     }
     while (!queue.empty()) {
-        auto *const state = queue.front();
+        auto* const state = queue.front();
         queue.pop();
         if (visited.find(state) != visited.end()) continue;
         visited.emplace(state);
@@ -72,8 +72,8 @@ std::vector<FSMState*> FSM::get_all_child_states(bool include_extra_state) const
         if (!state) continue;
         auto next_states = state->transitions();
         for (auto const& iter : next_states) {
-            auto *next_state = iter.second;
-            auto const *fsm = next_state->parent();
+            auto* next_state = iter.second;
+            auto const* fsm = next_state->parent();
             while (fsm && fsm != this) {
                 fsm = fsm->parent_fsm();
             }
@@ -124,11 +124,11 @@ void FSM::realize() {
     if (generator_->debug) {
         std::unordered_set<const FSM*> visited;
         for (auto const& state : states) {
-            const auto *fsm = state->parent();
+            const auto* fsm = state->parent();
             if (visited.find(fsm) != visited.end()) continue;
             auto fn_ln = fsm->fn_name_ln_;
             for (auto const& [name, info] : fn_ln) {
-                auto *s = fsm->get_state(name).get();
+                auto* s = fsm->get_state(name).get();
                 if (state_name_mapping.find(state) == state_name_mapping.end()) {
                     throw UserException(::format(
                         "Unable to find state name {0} from FSM {1}. Is it a disconnected state?",
@@ -228,7 +228,7 @@ std::shared_ptr<FunctionStmtBlock> FSM::get_func_def() {
     }
     auto ports = func->ports();
     for (auto const& [var_name, port] : ports) {
-        auto *var = name_mapping.at(var_name);
+        auto* var = name_mapping.at(var_name);
         func->add_stmt(var->assign(port, AssignmentType::Blocking));
     }
     return func;
@@ -269,7 +269,7 @@ void FSM::generate_state_transition(
             has_slide_through = true;
         }
         for (auto const& cond : vars) {
-            auto *next_fsm_state = transitions.at(cond);
+            auto* next_fsm_state = transitions.at(cond);
             if (!cond) {
                 // direct transition
                 auto stmt = get_next_state_stmt(enum_def, next_state, state, next_fsm_state,
@@ -374,8 +374,26 @@ void FSM::generate_output(Enum& enum_def, EnumVar& current_state,
     auto output_comb = generator_->combinational();
     auto output_case_comb = std::make_shared<SwitchStmt>(current_state.shared_from_this());
     auto states = get_all_child_states(false);
-    for (auto const& state : states) {
-        auto const& state_name = state_name_mapping.at(state);
+
+    // need to deal with fully specified case where default is the same as start state
+    // also default case
+    if (!is_2_power(enum_def.values.size())) {
+        states.emplace_back(nullptr);
+    }
+
+    for (auto* state : states) {
+        std::shared_ptr<EnumConst> state_enum;
+        std::string state_name;
+        if (state) {
+            state_name = state_name_mapping.at(state);
+            state_enum = enum_def.get_enum(state_name);
+        } else {
+            state_name = "default";
+            state_enum = nullptr;
+            // use default state outputs
+            state = start_state_.get();
+        }
+
         std::vector<std::shared_ptr<Stmt>> stmts;
         auto const& output_values = state->output_values();
         stmts.reserve(output_values.size());
@@ -386,7 +404,7 @@ void FSM::generate_output(Enum& enum_def, EnumVar& current_state,
         std::sort(vars.begin(), vars.end(),
                   [](auto& lhs, auto& rhs) { return lhs->name < rhs->name; });
         for (auto const& output_var : vars) {
-            auto *value = output_values.at(output_var);
+            auto* value = output_values.at(output_var);
             if (value && value != output_var) {
                 // value can be a nullptr
                 // users may use the same variable to indicate not changed
@@ -401,8 +419,7 @@ void FSM::generate_output(Enum& enum_def, EnumVar& current_state,
         }
         // add it to the case
         if (!stmts.empty()) {
-            auto& case_stmt =
-                output_case_comb->add_switch_case(enum_def.get_enum(state_name), stmts);
+            auto& case_stmt = output_case_comb->add_switch_case(state_enum, stmts);
             generator_->add_named_block(::format("{0}_{1}_Output", fsm_name_, state_name),
                                         case_stmt.as<ScopedStmtBlock>());
         }
@@ -515,7 +532,7 @@ std::string FSM::dot_graph() {
             return lhs->to_string() < rhs->to_string();
         });
         for (auto const& cond : conds) {
-            auto *next_state = transitions.at(cond);
+            auto* next_state = transitions.at(cond);
             if (cond) {
                 stream << indent
                        << ::format("{0}    ->  {1} [ label = \"{2}\" ];", state_name,
@@ -557,7 +574,7 @@ std::string FSM::output_table() {
     for (auto const& [state_name, state] : states_) {
         stream << state_name;
         for (auto const& output : outputs) {
-            auto *value = state->output_values().at(output);
+            auto* value = state->output_values().at(output);
             stream << "," << value->to_string();
         }
         stream << ::endl;
@@ -578,7 +595,7 @@ std::vector<FSM*> FSM::get_all_child_fsm() const {
     std::unordered_set<const FSM*> visited;
     queue.emplace(const_cast<FSM*>(this));
     while (!queue.empty()) {
-        auto *fsm = queue.front();
+        auto* fsm = queue.front();
         queue.pop();
         result.emplace_back(fsm);
         if (visited.find(fsm) != visited.end())
@@ -598,7 +615,7 @@ std::map<FSMState*, color::Color> get_state_color(const std::vector<FSMState*>&
     std::mt19937 gen(1);  // NOLINT
     std::uniform_real_distribution<double> dis(0, 1.0);
     for (auto const& state : states) {
-        auto fsm = state->parent();
+        auto const* fsm = state->parent();
         if (state_color.find(fsm) == state_color.end()) {
             // get a new color
             double h = dis(gen);
@@ -612,8 +629,8 @@ std::map<FSMState*, color::Color> get_state_color(const std::vector<FSMState*>&
     }
 
     // second pass the assign colors
-    for (const auto & state : states) {
-        const auto *fsm = state->parent();
+    for (const auto& state : states) {
+        const auto* fsm = state->parent();
         result.emplace(state, state_color.at(fsm));
     }
 
@@ -627,10 +644,10 @@ void FSMState::next(const std::shared_ptr<FSMState>& next_state, const std::shar
         throw UserException(
             ::format("Next state for {0}.{1} cannot be null", parent_->fsm_name(), name_));
     // making sure that it's part of the same fsm state
-    auto *parent = parent_;
-    auto *state_ptr = next_state.get();
+    auto* parent = parent_;
+    auto* state_ptr = next_state.get();
     while (parent->parent_fsm()) parent = parent->parent_fsm();
-    auto *fsm = next_state->parent_;
+    auto* fsm = next_state->parent_;
     while (fsm && fsm != parent) {
         fsm = fsm->parent_fsm();
     }
@@ -647,7 +664,7 @@ void FSMState::next(const std::shared_ptr<FSMState>& next_state, const std::shar
             // we have a slide through
             throw UserException("Unconditional transition has been assign to " + name_);
         }
-        auto *ptr = cond.get();
+        auto* ptr = cond.get();
         if (cond->width() != 1) throw VarException("Condition has to be a boolean value", {ptr});
 
         if (transitions_.find(ptr) != transitions_.end()) {
@@ -666,8 +683,8 @@ void FSMState::next(const std::shared_ptr<FSMState>& next_state, const std::shar
 
 void FSMState::output(const std::shared_ptr<Var>& output_var,
                       const std::shared_ptr<Var>& value_var) {
-    auto *output = output_var.get();
-    auto *value = value_var.get();
+    auto* output = output_var.get();
+    auto* value = value_var.get();
     if (output_values_.find(output) != output_values_.end()) {
         throw VarException(::format("{0} already has specified output"),
                            {output_values_.at(output)});
@@ -712,8 +729,8 @@ void FSMState::check_outputs() {
 
 std::string FSMState::handle_name() const {
     auto str = name_;
-    auto *current = parent_;
-    auto *p = parent_->parent_fsm();
+    auto* current = parent_;
+    auto* p = parent_->parent_fsm();
     while (p) {
         str = ::format("{0}_{1}", current->fsm_name(), str);
         current = p;
