@@ -21,13 +21,13 @@ using std::vector;
 
 namespace kratos {
 
-Generator Generator::from_verilog(Context *context, const std::string &src_file,
-                                  const std::string &top_name,
-                                  const std::vector<std::string> &lib_files,
-                                  const std::map<std::string, PortType> &port_types) {
+Generator &Generator::from_verilog(Context *context, const std::string &src_file,
+                                   const std::string &top_name,
+                                   const std::vector<std::string> &lib_files,
+                                   const std::map<std::string, PortType> &port_types) {
     if (!fs::exists(src_file)) throw UserException(::format("{0} does not exist", src_file));
 
-    Generator mod(context, top_name);
+    auto &mod = context->generator(top_name);
     // the src file will be treated a a lib file as well
     mod.lib_files_.reserve(1 + lib_files.size());
     mod.lib_files_.emplace_back(src_file);
@@ -103,6 +103,15 @@ Var &Generator::var(const std::string &var_name, uint32_t width, const std::vect
     return *p;
 }
 
+Var &Generator::var(const Var &v, const std::string &var_name) {
+    auto &v_ = var(var_name, v.var_width(), v.size(), v.is_signed());
+    v_.set_explicit_array(v.explicit_array());
+    v_.set_is_packed(v.is_packed());
+    // need to copy other definition over
+    v_.copy_meta_data(&v_, true);
+    return v_;
+}
+
 std::shared_ptr<Var> Generator::get_var(const std::string &var_name) {
     if (vars_.find(var_name) == vars_.end()) return nullptr;
     return vars_.at(var_name);
@@ -157,7 +166,7 @@ Port &Generator::port(const PortPackedStruct &port, const std::string &port_name
         throw VarException(::format("{0} already exists in {1}", port_name, name),
                            {vars_.at(port_name).get()});
     auto p = std::make_shared<PortPackedStruct>(this, port.port_direction(), port_name,
-                                                port.packed_struct(), port.size());
+                                                port.packed_struct());
     vars_.emplace(port_name, p);
     ports_.emplace(port_name);
 
@@ -370,6 +379,7 @@ void Generator::add_function(const std::shared_ptr<FunctionStmtBlock> &func) {
 }
 
 IRNode *Generator::get_child(uint64_t index) {
+    if (is_external_) return nullptr;
     if (index < stmts_count()) {
         return stmts_[index].get();
     } else if (index < stmts_count() + funcs_.size()) {
@@ -453,6 +463,7 @@ void Generator::remove_child_generator(const std::shared_ptr<Generator> &child)
 
 std::vector<std::shared_ptr<Generator>> Generator::get_child_generators() {
     std::vector<std::shared_ptr<Generator>> result;
+    if (is_external_) return result;
     result.reserve(children_.size());
     for (auto const &n : children_names_) {
         result.emplace_back(children_.at(n));
@@ -575,6 +586,23 @@ void Generator::rename_var(const std::string &old_name, const std::string &new_n
     vars_.insert(std::move(handle));
 }
 
+void Generator::reindex_vars() {
+    // this is a little bit expensive in terms of computation
+    std::map<std::string, std::shared_ptr<Var>> vars;
+    std::set<std::string> ports;
+
+    for (auto const &[n_, var] : vars_) {
+        auto const &name_ = var->name;
+        vars.emplace(name_, var);
+        if (var->type() == VarType::PortIO) {
+            ports_.emplace(name_);
+        }
+    }
+
+    vars_ = vars;
+    ports_ = ports;
+}
+
 void Generator::add_call_var(const std::shared_ptr<FunctionCallVar> &var) {
     calls_.emplace(var);
     var->set_generator(this);
@@ -634,6 +662,11 @@ std::shared_ptr<InterfaceRef> Generator::interface(const std::shared_ptr<IDefini
     return ref;
 }
 
+std::shared_ptr<InterfaceRef> Generator::get_interface(const std::string &interface_name) const {
+    return interfaces_.find(interface_name) != interfaces_.end() ? interfaces_.at(interface_name)
+                                                                 : nullptr;
+}
+
 std::shared_ptr<SequentialStmtBlock> Generator::sequential() {
     auto stmt = std::make_shared<SequentialStmtBlock>();
     add_stmt(stmt);
@@ -652,6 +685,12 @@ std::shared_ptr<InitialStmtBlock> Generator::initial() {
     return stmt;
 }
 
+std::shared_ptr<FinalStmtBlock> Generator::final() {
+    auto stmt = std::make_shared<FinalStmtBlock>();
+    add_stmt(stmt);
+    return stmt;
+}
+
 std::shared_ptr<LatchStmtBlock> Generator::latch() {
     auto stmt = std::make_shared<LatchStmtBlock>();
     add_stmt(stmt);
@@ -684,11 +723,11 @@ std::pair<bool, bool> correct_port_direction(Port *port1, Port *port2, Generator
         bool interface_port_wire = false;
         if (port1->is_interface() && !port2->is_interface()) {
             auto *port_i = reinterpret_cast<InterfacePort *>(port1);
-            auto *it = port_i->interface();
+            const auto *it = port_i->interface();
             interface_port_wire = !it->is_port();
         } else if (port2->is_interface() && !port1->is_interface()) {
             auto *port_i = reinterpret_cast<InterfacePort *>(port2);
-            auto *it = port_i->interface();
+            const auto *it = port_i->interface();
             interface_port_wire = !it->is_port();
         }
         check_direction(port1, port2, interface_port_wire);
@@ -752,7 +791,7 @@ std::pair<bool, bool> Generator::correct_wire_direction(const std::shared_ptr<Va
             bool flip_dir = false;
             if (port1->is_interface()) {
                 auto *port_i = reinterpret_cast<InterfacePort *>(port1);
-                auto *ref = port_i->interface();
+                const auto *ref = port_i->interface();
                 flip_dir = !(ref->is_port() && root2->generator() == this);
             }
             return {!flip_dir ? port1->port_direction() == PortDirection::Out
@@ -774,7 +813,7 @@ std::pair<bool, bool> Generator::correct_wire_direction(const std::shared_ptr<Va
             bool flip_dir = false;
             if (port2->is_interface()) {
                 auto *port_i = reinterpret_cast<InterfacePort *>(port2);
-                auto *ref = port_i->interface();
+                const auto *ref = port_i->interface();
                 flip_dir = !(ref->is_port() && root1->generator() == this);
             }
             return {!flip_dir ? port2->port_direction() == PortDirection::In
@@ -843,13 +882,35 @@ void Generator::wire_interface(const std::shared_ptr<InterfaceRef> &inst1,
 
 void Generator::wire(Var &left, Var &right) { add_stmt(left.assign(right)); }
 
+void Generator::unwire(Var &var1, Var &var2) {
+    // brute force search matching statement
+    std::shared_ptr<Stmt> target = nullptr;
+    for (auto const &stmt : stmts_) {
+        if (stmt->type() == StatementType::Assign) {
+            auto assign_stmt = stmt->as<AssignStmt>();
+            if ((assign_stmt->left() == &var1 && assign_stmt->right() == &var2) ||
+                (assign_stmt->right() == &var1 && assign_stmt->left() == &var2)) {
+                target = assign_stmt;
+                break;
+            }
+        }
+    }
+    if (target) {
+        auto assign_stmt = target->as<AssignStmt>();
+        assign_stmt->left()->remove_source(assign_stmt);
+        assign_stmt->right()->remove_sink(assign_stmt);
+        remove_stmt(target);
+    }
+}
+
 std::shared_ptr<Generator> Generator::clone() {
     auto generator = std::make_shared<Generator>(context_, name);
     auto port_names = get_port_names();
     for (auto const &port_name : port_names) {
         auto port = get_port(port_name);
-        generator->port(port->port_direction(), port_name, port->width(), port->size(),
-                        port->port_type(), port->is_signed());
+        auto &p = generator->port(port->port_direction(), port_name, port->var_width(),
+                                  port->size(), port->port_type(), port->is_signed());
+        p.set_is_packed(port->is_packed());
     }
     // also parameters
     for (auto const &[param_name, param] : params_) {
@@ -871,17 +932,18 @@ void Generator::accept(IRVisitor *visitor) {
 }
 
 PortPackedStruct &Generator::port_packed(PortDirection direction, const std::string &port_name,
-                                         const PackedStruct &packed_struct_) {
+                                         const std::shared_ptr<PackedStruct> &packed_struct_) {
     return port_packed(direction, port_name, packed_struct_, 1);
 }
 
 PortPackedStruct &Generator::port_packed(PortDirection direction, const std::string &port_name,
-                                         const PackedStruct &packed_struct_, uint32_t size) {
+                                         const std::shared_ptr<PackedStruct> &packed_struct_,
+                                         uint32_t size) {
     return port_packed(direction, port_name, packed_struct_, std::vector<uint32_t>{size});
 }
 
 PortPackedStruct &Generator::port_packed(PortDirection direction, const std::string &port_name,
-                                         const PackedStruct &packed_struct_,
+                                         const std::shared_ptr<PackedStruct> &packed_struct_,
                                          const std::vector<uint32_t> &size) {
     if (ports_.find(port_name) != ports_.end())
         throw VarException(::format("{0} already exists in {1}", port_name, name),
@@ -893,17 +955,18 @@ PortPackedStruct &Generator::port_packed(PortDirection direction, const std::str
 }
 
 VarPackedStruct &Generator::var_packed(const std::string &var_name,
-                                       const kratos::PackedStruct &packed_struct_) {
+                                       const std::shared_ptr<PackedStruct> &packed_struct_) {
     return var_packed(var_name, packed_struct_, std::vector<uint32_t>{1});
 }
 
 VarPackedStruct &Generator::var_packed(const std::string &var_name,
-                                       const PackedStruct &packed_struct_, uint32_t size) {
+                                       const std::shared_ptr<PackedStruct> &packed_struct_,
+                                       uint32_t size) {
     return var_packed(var_name, packed_struct_, std::vector<uint32_t>{size});
 }
 
 VarPackedStruct &Generator::var_packed(const std::string &var_name,
-                                       const PackedStruct &packed_struct_,
+                                       const std::shared_ptr<PackedStruct> &packed_struct_,
                                        const std::vector<uint32_t> &size) {
     if (vars_.find(var_name) != vars_.end())
         throw VarException(::format("{0} already exists in {1}", var_name, name),
@@ -1065,10 +1128,10 @@ void Generator::remove_var(const std::string &var_name) {
     }
     auto var = vars_.at(var_name);
     if (!var->sources().empty()) {
-        throw UserException(::format("{0} still has source connection(s)"));
+        throw UserException(::format("{0} still has source connection(s)", var->name));
     }
     if (!var->sinks().empty()) {
-        throw UserException(::format("{0} still has sink connection(s)"));
+        throw UserException(::format("{0} still has sink connection(s)", var->name));
     }
 
     vars_.erase(var_name);
@@ -1079,6 +1142,15 @@ std::shared_ptr<StmtBlock> Generator::get_named_block(const std::string &block_n
     return named_blocks_.at(block_name);
 }
 
+std::optional<std::string> Generator::get_block_name(const Stmt *stmt) const {
+    for (auto const &[name, s] : named_blocks_) {
+        if (s.get() == stmt) {
+            return name;
+        }
+    }
+    return std::nullopt;
+}
+
 void Generator::add_named_block(const std::string &block_name,
                                 const std::shared_ptr<StmtBlock> &block) {
     if (has_named_block(block_name))
@@ -1098,21 +1170,34 @@ std::string Generator::handle_name() const { return handle_name(false); }
 std::string Generator::handle_name(bool ignore_top) const {
     // this is used to identify the generator from the top level
     std::string result = instance_name;
-    auto *parent = parent_generator_;
-    if (ignore_top) {
-        std::vector<std::string> values;
-        values.emplace_back(instance_name);
-        while (parent != nullptr) {
-            values.emplace(values.begin(), parent->instance_name);
-            parent = parent->parent_generator_;
-        }
-        result = string::join(values.begin() + 1, values.end(), ".");
-    } else {
-        while (parent != nullptr) {
-            result = ::format("{0}.{1}", parent->instance_name, result);
-            parent = parent->parent_generator_;
+    auto const *current = this;
+    std::vector<std::string> values;
+    while (current != nullptr) {
+        // need to check if we are in a gen block
+        if (current->parent_generator_ && current->instantiation_stmt_ &&
+            current->instantiation_stmt_->parent()->ir_node_kind() == IRNodeKind::StmtKind) {
+            // need to find out the label
+            // need two parents since the previous parent is a stmt block
+            auto *for_ = reinterpret_cast<ForStmt *>(current->instantiation_stmt_->parent());
+            auto label = current->parent_generator_->get_block_name(for_->get_loop_body().get());
+            // need to find out the index
+            auto index = for_->genvar_index(current->instantiation_stmt_->shared_from_this());
+            if (!label || !index) {
+                throw InternalException("Invalid state of genvar instance array");
+            }
+
+            auto n = ::format("{0}[{1}]", current->instance_name, *index);
+            values.emplace(values.begin(), n);
+            values.emplace(values.begin(), *label);
+        } else {
+            values.emplace(values.begin(), current->instance_name);
         }
+
+        current = current->parent_generator_;
     }
+    auto starting = ignore_top ? 1u : 0u;
+    result = string::join(values.begin() + starting, values.end(), ".");
+
     return result;
 }
 
