@@ -42,6 +42,10 @@ Generator *Stmt::generator_parent() const {
     return dynamic_cast<Generator *>(p);
 }
 
+void Stmt::set_scope_context(const std::map<std::string, std::pair<bool, std::string>> &context) {
+    scope_context_ = context;
+}
+
 void Stmt::add_scope_variable(const std::string &name, const std::string &value, bool is_var,
                               bool override) {
     if (override || scope_context_.find(name) == scope_context_.end()) {
@@ -231,16 +235,20 @@ void IfStmt::remove_stmt(const std::shared_ptr<kratos::Stmt> &stmt) {
 }
 
 void IfStmt::set_then(const std::shared_ptr<ScopedStmtBlock> &stmt) {
-    then_body_->clear();
-    for (auto &s : *stmt) {
-        then_body_->add_stmt(s);
+    if (stmt != then_body_) {
+        then_body_->clear();
+        for (auto &s : *stmt) {
+            then_body_->add_stmt(s);
+        }
     }
 }
 
 void IfStmt::set_else(const std::shared_ptr<ScopedStmtBlock> &stmt) {
-    else_body_->clear();
-    for (auto &s : *stmt) {
-        else_body_->add_stmt(s);
+    if (else_body_ != stmt) {
+        else_body_->clear();
+        for (auto &s : *stmt) {
+            else_body_->add_stmt(s);
+        }
     }
 }
 
@@ -302,8 +310,23 @@ ForStmt::ForStmt(const std::string &iter_var_name, int64_t start, int64_t end, i
 }
 
 void ForStmt::set_parent(IRNode *node) {
-    if (node->ir_node_kind() != IRNodeKind::StmtKind)
-        throw UserException("For loop can only be added to statement body");
+    if (node->ir_node_kind() != IRNodeKind::StmtKind) {
+        // make sure it is genvar and all it's statements is instantiation statement
+        bool all_instantiation = std::all_of(
+            loop_body_->begin(), loop_body_->end(), [](const std::shared_ptr<Stmt> &stmt) {
+                return stmt->type() == StatementType::ModuleInstantiation;
+            });
+        // make sure it's genvar marked
+        bool is_genvar = iter_->is_gen_var();
+        if (!is_genvar || !all_instantiation || node->ir_node_kind() != IRNodeKind::GeneratorKind) {
+            throw UserException("For loop can only be added to statement body");
+        } else {
+            auto *gen = reinterpret_cast<Generator *>(node);
+            iter_->set_generator(gen);
+            Stmt::set_parent(node);
+            return;
+        }
+    }
     auto *stmt = reinterpret_cast<Stmt *>(node);
     auto *gen = stmt->generator_parent();
     if (gen) {
@@ -331,6 +354,15 @@ std::shared_ptr<Stmt> ForStmt::clone() const {
 
 void ForStmt::clear() { loop_body_->clear(); }
 
+std::optional<uint64_t> ForStmt::genvar_index(const std::shared_ptr<Stmt> &stmt) {
+    auto pos = std::find(gen_var_stmts_.begin(), gen_var_stmts_.end(), stmt);
+    if (pos == gen_var_stmts_.end()) {
+        return std::nullopt;
+    } else {
+        return pos - gen_var_stmts_.begin();
+    }
+}
+
 StmtBlock::StmtBlock(StatementBlockType type) : Stmt(StatementType::Block), block_type_(type) {}
 
 void StmtBlock::add_stmt(const std::shared_ptr<Stmt> &stmt) {
@@ -454,21 +486,28 @@ void SwitchStmt::set_parent(IRNode *parent) {
 ScopedStmtBlock &SwitchStmt::add_switch_case(const std::shared_ptr<Const> &switch_case,
                                              const std::shared_ptr<Stmt> &stmt) {
     stmt->set_parent(this);
-    if (body_.find(switch_case) == body_.end()) {
-        auto block = std::make_shared<ScopedStmtBlock>();
-        block->set_parent(this);
-        body_.emplace(switch_case, block);
+    std::shared_ptr<ScopedStmtBlock> switch_stmt;
+    for (auto const &[c, s] : body_) {
+        if ((!switch_case && !c) || (switch_case && c && c->value() == switch_case->value())) {
+            switch_stmt = s;
+            break;
+        }
+    }
+    if (!switch_stmt) {
+        switch_stmt = std::make_shared<ScopedStmtBlock>();
+        switch_stmt->set_parent(this);
+        body_.emplace(switch_case, switch_stmt);
     }
     if (stmt->type() == StatementType::Block) {
         // merge the block
         auto blk = stmt->as<StmtBlock>();
         for (auto const &s : *blk) {
-            body_[switch_case]->add_stmt(s);
+            switch_stmt->add_stmt(s);
         }
     } else {
-        body_[switch_case]->add_stmt(stmt);
+        switch_stmt->add_stmt(stmt);
     }
-    return *body_[switch_case];
+    return *switch_stmt;
 }
 
 ScopedStmtBlock &SwitchStmt::add_switch_case(const std::shared_ptr<Const> &switch_case,
@@ -512,9 +551,19 @@ IRNode *SwitchStmt::get_child(uint64_t index) {
     if (index == 0) {
         return target_.get();
     } else {
+        // need to make it deterministic
         index--;
+        std::vector<std::pair<uint64_t, IRNode *>> cases;
+        cases.reserve(body_.size());
         for (auto const &iter : body_) {
-            if (index-- == 0) return iter.second.get();
+            cases.emplace_back(std::make_pair(
+                iter.first ? iter.first->value() : std::numeric_limits<uint64_t>::max(),
+                iter.second.get()));
+        }
+        std::sort(cases.begin(), cases.end(),
+                  [](auto const &a, auto const &b) { return a.first < b.first; });
+        if (index < cases.size()) {
+            return cases[index].second;
         }
         return nullptr;
     }
@@ -884,4 +933,46 @@ std::shared_ptr<Stmt> RawStringStmt::clone() const {
     return stmt;
 }
 
+AuxiliaryStmt::AuxiliaryStmt(AuxiliaryType type) : Stmt(StatementType::Auxiliary), type_(type) {}
+
+EventActionType operator|=(EventActionType &lhs, EventActionType rhs) {
+    lhs =
+        static_cast<EventActionType>(static_cast<std::underlying_type<EventActionType>::type>(lhs) |
+                                     static_cast<std::underlying_type<EventActionType>::type>(rhs));
+    return lhs;
+}
+
+EventTracingStmt::EventTracingStmt(std::string name)
+    : AuxiliaryStmt(AuxiliaryType::EventTracing), event_name_(std::move(name)) {}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::terminates() {
+    action_type_ |= EventActionType::End;
+    return as<EventTracingStmt>();
+}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::terminates(
+    const std::string &transaction_name) {
+    return belongs(transaction_name)->terminates();
+}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::belongs(const std::string &transaction_name) {
+    transaction_ = transaction_name;
+    return as<EventTracingStmt>();
+}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::starts() {
+    action_type_ |= EventActionType::Start;
+    return as<EventTracingStmt>();
+}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::starts(const std::string &transaction_name) {
+    return belongs(transaction_name)->starts();
+}
+
+std::shared_ptr<EventTracingStmt> EventTracingStmt::matches(const std::string &field_name,
+                                                            const std::shared_ptr<Var> &var) {
+    match_values_.emplace(field_name, var);
+    return as<EventTracingStmt>();
+}
+
 }  // namespace kratos
