@@ -26,6 +26,9 @@ TEST(generator, load) {  // NOLINT
     EXPECT_EQ(mod.get_port("a")->port_type(), PortType::Clock);
     ASSERT_ANY_THROW(
         Generator::from_verilog(&c, "module1.sv", "module1", {}, {{"aa", PortType::Clock}}));
+    mod = (Generator::from_verilog(&c, "module1.sv", "module4", {}, {}));
+    EXPECT_TRUE(mod.get_port("input_port") != nullptr);
+    EXPECT_TRUE(mod.get_port("output_port") != nullptr);
 }
 
 TEST(generator, port) {  // NOLINT
@@ -1544,7 +1547,9 @@ TEST(codegen, param_size_single_module) { // NOLINT
 
     auto &param = mod.parameter("P", 5);
     auto &in = mod.port(PortDirection::In, "in", 5, 3);
+    in.set_is_packed(false);
     auto &in2 = mod.port(PortDirection::In, "in2", 6, {3, 2});
+    in2.set_is_packed(false);
     // create a slice
     in[2];
     // this should be an exception
@@ -1570,11 +1575,13 @@ TEST(codegen, param_size_nested_module) {   // NOLINT
     // c_p has to have value
     c_p.set_value(1);
     auto &c_in = child.port(PortDirection::In, "in", 16, 32);
+    c_in.set_is_packed(false);
     c_in.set_size_param(0, &c_p);
 
     auto &parent = c.generator("parent");
     auto &p_p = parent.parameter("P", 32);
     auto &p_in = parent.port(PortDirection::In, "in", 16, 32);
+    p_in.set_is_packed(false);
     p_p.set_value(2);
     p_in.set_size_param(0, &p_p);
 
@@ -1600,6 +1607,7 @@ TEST(codegen, copy_port_definition) {   // NOLINT
     // c_p has to have value
     c_p.set_value(1);
     auto &c_in = child.port(PortDirection::In, "in", 16, 32);
+    c_in.set_is_packed(false);
     c_in.set_size_param(0, &(c_p * constant(2, 16)));
 
     auto &parent = c.generator("parent");
@@ -1620,4 +1628,19 @@ TEST(codegen, copy_port_definition) {   // NOLINT
     auto src = generate_verilog(&parent);
     auto mod_src = src.at("parent");
     EXPECT_NE(mod_src.find("input logic [15:0] in [(P * 32'h2)-1:0]"), std::string::npos);
+}
+
+TEST(generator, unwire) {   // NOLINT
+    Context c;
+    auto &mod = c.generator("mod");
+    auto &a = mod.var("a", 1);
+    auto &b = mod.var("b", 1);
+
+    mod.unwire(b, a);
+    EXPECT_EQ(mod.stmts_count(), 0);
+
+    mod.add_stmt(a.assign(b));
+    EXPECT_EQ(mod.stmts_count(), 1);
+    mod.unwire(b, a);
+    EXPECT_EQ(mod.stmts_count(), 0);
 }
\ No newline at end of file
