@@ -1132,27 +1132,8 @@ def test_symbol_table():
     assert len(table[mod]) == 5
 
 
-def test_breakpoint(check_gold):
-    from _kratos.passes import extract_debug_break_points, \
-        hash_generators_sequential
-    mod = Generator("mod", True)
-    comb = mod.combinational()
-    stmt0 = mod.output("out", 1).assign(mod.input("in", 1))
-    comb.add_stmt(stmt0)
-    stmt1 = mod.var("val", 1).assign(mod.ports["in"])
-    comb.add_stmt(stmt1)
-
-    hash_generators_sequential(mod.internal_generator)
-    enable_runtime_debug(mod)
-    table = extract_debug_break_points(mod.internal_generator)
-    assert len(table) == 2
-    assert table[stmt0] == 0
-    assert table[stmt1] == 1
-    check_gold(mod, "test_breakpoint")
-
-
 def test_cast():
-    from kratos.util import clock
+    from kratos.util import clock, clock_en
 
     class Mod1(Generator):
         def __init__(self):
@@ -1176,10 +1157,20 @@ def test_cast():
         def __init__(self):
             super().__init__("mod")
             self.v = self.var("v", 1)
+            clk = clock(self.v)
+            cn_var = self.var("cn", 1)
+            cn = self.output("cn_out", 1, PortType.ClockEnable)
+            self.add_stmt(cn.assign(clock_en(cn_var)))
             # only procedural allowed
             seq = self.sequential((posedge, clock(self.v)))
             seq.add_stmt(self.output("out", 1).assign(const(1, 1)))
 
+            @always_ff((posedge, clk))
+            def fn():
+                self.v = 1
+
+            self.add_always(fn)
+
     mod = Mod2()
     verilog(mod)
 
@@ -1486,12 +1477,18 @@ def test_wrapper_flatten_generator(check_gold):
     mod = Generator("mod")
     a = mod.input("a", 4, size=[3, 2])
     b = mod.output("b", 4, size=[3, 2])
+    attr = kratos.Attribute()
+    attr.value_str = "a"
+    a.add_attribute(attr)
     mod.wire(a, b)
     from _kratos import create_wrapper_flatten
     wrapper = create_wrapper_flatten(mod.internal_generator, "wrapper")
     wrapper = Generator(wrapper.name, internal_generator=wrapper)
     check_gold(wrapper, gold_name="test_wrapper_flatten_generator",
                optimize_passthrough=False)
+    a_0_0 = wrapper.ports.a_0_0
+    attrs = a_0_0.find_attribute(lambda x: x.value_str == "a")
+    assert len(attrs) == 1
 
 
 def test_register_extraction():
@@ -1898,7 +1895,7 @@ def test_ports_vars_iter():
         vs.add(var_name)
     assert len(vs) == 3
     params = set()
-    for param_name in mod.params:
+    for param_name, _ in mod.params:
         params.add(param_name)
     assert len(params) == 2
 
@@ -1928,6 +1925,17 @@ def test_port_from_def():
     assert p1.width == p2.width
 
 
+def test_var_from_def():
+    mod = Generator("mod")
+    v1 = mod.var("v1", 2, size=[2, 3])
+    v2 = mod.var_from_def(v1, "v2")
+    assert v2.width == v1.width
+    assert v2.name == "v2"
+    p1 = mod.input("p1", 42)
+    v3 = mod.var_from_def(p1, "v3")
+    assert v3.width == 42
+
+
 def test_raw_import():
     mod = Generator("mod")
     mod.internal_generator.add_raw_import("pkg_name")
@@ -2097,5 +2105,229 @@ def test_clog2_var():
     assert str(f) == "c[d]"
 
 
+def test_ssa_transform(check_gold):
+    mod = Generator("mod", debug=True)
+    a = mod.var("a", 4)
+    b = mod.var("b", 4)
+    c = mod.output("c", 4)
+
+    @always_comb
+    def func():
+        a = 1
+        a = 2
+        if a == 2:
+            a = b + a
+        if a == 3:
+            b = 2
+        else:
+            if a == 4:
+                b = 3
+            else:
+                b = 4
+                # this is not a latch
+                a = 5
+        c = a
+
+    mod.add_always(func, ssa_transform=True)
+    check_gold(mod, "test_ssa_transform", ssa_transform=True)
+    # check if the local variable mapping is fixed
+    # assign a_5 = (a_3 == 4'h4) ? a_3: a_4;
+    # which corresponds to a = 5
+    # notice that a should be pointing to a = b + a, since it's the last
+    # time a gets assigned
+    stmt = mod.get_stmt_by_index(6)
+    scope = stmt.scope_context
+    is_var, a_mapping = scope["a"]
+    assert is_var
+    # this is assign a_2 = b + a_1;
+    stmt = mod.get_stmt_by_index(5)
+    assert str(a_mapping) == str(stmt.left)
+
+    # test enable table extraction
+    from _kratos.passes import compute_enable_condition
+    enable_map = compute_enable_condition(mod.internal_generator)
+    assert len(enable_map) > 5
+
+
+def test_enable_condition_always_ff():
+    mod = Generator("mod")
+    a = mod.var("a", 4)
+    b = mod.var("b", 1)
+    clk = mod.clock("clk")
+
+    @always_ff((posedge, clk))
+    def logic():
+        if b:
+            a = 0
+        else:
+            a = 1
+
+    mod.add_always(logic)
+    from _kratos.passes import compute_enable_condition
+    enable_map = compute_enable_condition(mod.internal_generator)
+    assert len(enable_map) == 2
+
+
+def test_var_rename():
+    import _kratos
+    mod = Generator("mod")
+    a = mod.var("a", 1)
+    b = mod.input("b", 1)
+    a.rename("c")
+    b.rename("d")
+    assert "c" in mod.vars
+    assert "a" not in mod.vars
+    assert "c" == a.name
+    try:
+        a.name = "d"
+        assert False
+    except _kratos.exception.UserException:
+        pass
+
+
+def test_merge_const_port_assignment():
+    mod = Generator("mod")
+    a = mod.var("a", 1)
+    child = Generator("child")
+    b = child.input("b", 1)
+    mod.add_child("child", child, b=a)
+    mod.add_stmt(a.assign(1))
+    v = verilog(mod)["mod"]
+    assert ".b(1'h1)" in v
+
+
+def test_gen_inst_lift(check_gold):
+    num_inst = 4
+    parent = Generator("parent")
+    clk = parent.clock("clk")
+    a_array = parent.var("a", 1, size=4)
+    b_array = parent.var("b", 1, size=4)
+    children = []
+
+    for i in range(num_inst):
+        child = Generator("child")
+        child.clock("clk")
+        a = child.input("a", 1)
+        b = child.output("b", 1)
+        child.wire(a, b)
+        parent.add_child(f"child_{i}", child,
+                         clk=clk,
+                         a=a_array[i],
+                         b=b_array[i])
+        children.append(child)
+
+    check_gold(parent, "test_gen_inst_lift", lift_genvar_instances=True)
+    name = children[1].internal_generator.handle_name()
+    assert name == "parent.child.inst[1]"
+
+    # another one that will fail the genvar test
+    Generator.clear_context()
+    num_inst = 4
+    parent = Generator("parent")
+    clk = parent.clock("clk")
+    a_array = parent.var("a", 1, size=2)
+
+    for i in range(2):
+        child = Generator("child")
+        child.clock("clk")
+        child.input("a", 1)
+        parent.add_child(f"child_{i}", child,
+                         clk=clk,
+                         a=a_array[0])
+    src = verilog(parent, lift_genvar_instances=True)["parent"]
+    assert "genvar" not in src
+
+
+def test_add_child_interface_port_wiring(check_gold):
+    from kratos import Interface
+    mod = Generator("mod")
+
+    class ConfigInterface(Interface):
+        def __init__(self):
+            Interface.__init__(self, "Config")
+            self.input("read", 1, 1)
+            self.output("write", 1, 1)
+
+    interface = ConfigInterface()
+    i1 = mod.interface(interface, "bus")
+    # wire local variables to the interface
+    read = mod.var("read", 1)
+    write = mod.var("write", 1)
+    mod.wire(i1.read, read)
+    mod.wire(write, i1.write)
+
+    # create child module and use the interface as port
+    child = Generator("child")
+    i2 = child.interface(interface, "bus2", True)
+    child.wire(i2.write, i2.read)
+
+    # wire the interface using add_child
+    mod.add_child("child", child, bus2=i1)
+    check_gold(mod, "test_add_child_interface_port_wiring",
+               optimize_passthrough=False)
+
+
+def test_auto_insert_clk_gate_skip():
+    from _kratos.passes import auto_insert_clock_enable
+
+    class Mod(Generator):
+        def __init__(self, add_self, add_always):
+            super(Mod, self).__init__("mod")
+            clk = self.clock("clk")
+            a = self.var("a", 1)
+            self.clock_en("clk_en")
+
+            @always_ff((posedge, clk))
+            def code():
+                a = 1
+
+            b = self.add_always(code)
+            if add_always:
+                b.add_attribute("dont_touch")
+            if add_self:
+                self.add_attribute("dont_touch")
+    m = Mod(False, False)
+    auto_insert_clock_enable(m.internal_generator)
+    src = verilog(m)["mod"]
+    assert "if (clk_en)" in src
+    m = Mod(False, True)
+    auto_insert_clock_enable(m.internal_generator)
+    src = verilog(m)["mod"]
+    assert "if (clk_en)" not in src
+    m = Mod(True, False)
+    auto_insert_clock_enable(m.internal_generator)
+    src = verilog(m)["mod"]
+    assert "if (clk_en)" not in src
+
+
+def test_final_block():
+    from kratos import final
+
+    mod = Generator("mod")
+    a = mod.var("a", 1)
+
+    @final
+    def code():
+        a = 1
+
+    mod.add_always(code)
+    src = verilog(mod)["mod"]
+    assert "final begin\n  a = 1'h1;\nend" in src
+
+
+def test_struct_of_struct(check_gold):
+    mod = Generator("mod")
+    struct1 = PackedStruct("struct1")
+    struct1.add_attribute("value1", 32)
+    struct0 = PackedStruct("struct0")
+    struct0.add_attribute("value2", struct1)
+    v = mod.var("v", struct0)
+    v_array = mod.var("v_array", struct0, size=4)
+    mod.add_stmt(v["value2"]["value1"].assign(1))
+    mod.add_stmt(v_array[0]["value2"]["value1"].assign(1))
+    check_gold(mod, "test_struct_of_struct")
+
+
 if __name__ == "__main__":
-    test_clog2_var()
+    from conftest import check_gold_fn
+    test_struct_of_struct(check_gold_fn)
